
import logging
logger = logging.getLogger(__name__)

from termcolor import cprint, colored

from ... import types, common

elf_types = {
    0x0: 'NONE',
    0x1: 'REL',
    0x2: 'EXEC',
    0x3: 'DYN',
    0x4: 'CORE',
    0xfe00: 'LOOS',
    0xfeff: 'HIOS',
    0xff00: 'LOPROC',
    0xffff: 'HIPROC',
}

class ELF(object):

    def __init__(self, process, module):
        """Parses the ELF in memory.

        Args:
            process (Process): revenge Process object
            module (Module): Module object to start parsing at
        """

        self._process = process
        self.module = module
        self.bits = None

        assert self._process.memory[self.module.base:self.module.base+4].bytes == b'\x7fELF', "This does not appear to be an ELF file..."

    @property
    def program_headers(self):
        """Program headers object."""
        return ProgramHeaders(self._process, self)

    @property
    def section_headers(self):
        """Section headers object."""
        return SectionHeaders(self._process, self)

    @property
    def bits(self):
        """int: How many bits is this ELF?"""

        if self.__bits is None:
            self.__bits = 32 if self._process.memory[self.module.base+0x4].int8 == 1 else 64

        return self.__bits

    @bits.setter
    def bits(self, bits):
        self.__bits = bits

    @property
    def entry(self):
        """int: ELF entrypoint. Rebased"""
        ret = self._process.memory[self.module.base+0x18].pointer

        if self.type_str == 'DYN':
            ret = types.Pointer(ret + self.module.base)

        return ret

    @property
    def phoff(self):
        """int: Program header offset."""
        if self.bits == 32:
            ret = self._process.memory[self.module.base+0x1C].pointer
        else:
            ret = self._process.memory[self.module.base+0x20].pointer
            
        return types.Pointer(ret + self.module.base)

    @property
    def phnum(self):
        """int: Number of program headers."""
        if self.bits == 32:
            return self._process.memory[self.module.base+0x2C].uint16
        else:
            return self._process.memory[self.module.base+0x38].uint16

    @property
    def phentsize(self):
        """int: Size of program headers."""
        if self.bits == 32:
            return self._process.memory[self.module.base+0x2A].uint16
        else:
            return self._process.memory[self.module.base+0x36].uint16

    @property
    def flags(self):
        """int: Flags"""
        if self.bits == 32:
            return self._process.memory[self.module.base+0x24].uint32
        else:
            return self._process.memory[self.module.base+0x30].uint32

    @property
    def ehsize(self):
        """int: Size of this header."""
        if self.bits == 32:
            return self._process.memory[self.module.base+0x28].uint16
        else:
            return self._process.memory[self.module.base+0x34].uint16

    @property
    def shentsize(self):
        """int: Size of a section header entry"""
        if self.bits == 32:
            return self._process.memory[self.module.base+0x2e].uint16
        else:
            return self._process.memory[self.module.base+0x3a].uint16

    @property
    def shnum(self):
        """int: Number of section header entries."""
        if self.bits == 32:
            return self._process.memory[self.module.base+0x30].uint16
        else:
            return self._process.memory[self.module.base+0x3c].uint16

    @property
    def shstrndx(self):
        """int: Section header index that contains the section names."""
        if self.bits == 32:
            return self._process.memory[self.module.base+0x32].uint16
        else:
            return self._process.memory[self.module.base+0x3e].uint16

    @property
    def shoff(self):
        """int: Section Header offset"""
        if self.bits == 32:
            ret = self._process.memory[self.module.base+0x20].pointer
        else:
            ret = self._process.memory[self.module.base+0x28].pointer

        return types.Pointer(ret + self.module.base)

    @property
    def type(self):
        """int: Type of this binary."""

        try:
            return self.__type
        except AttributeError:
            self.__type = self._process.memory[self.module.base+0x10].uint16
            return self.__type

    @property
    def type_str(self):
        """str: Type of this binary."""
        return elf_types[self.type]
        

from .program_headers import ProgramHeaders
from .section_headers import SectionHeaders




import logging
logger = logging.getLogger(__name__)

from prettytable import PrettyTable
import typing

from . import common

class Thread(object):

    def __init__(self, process, info):
        self._process = process
        self._info = info
        self.context = Context(self._process, **self._info['context'])

    def __repr__(self):
        attrs = ['Thread', hex(self.id), '@', hex(self.pc), self.state, self.module]
        if self.trace is not None:
            attrs.append('tracing')
        return "<{}>".format(' '.join(attrs))

    def __getattr__(self, elm):
        return common.auto_int(self._info['context'][elm])

    def __str__(self):

        table = PrettyTable(['attr', 'value'])

        table.add_row(['TID', str(self.id)])
        table.add_row(["State", self.state])
        table.add_row(["Module", self.module])
        table.add_row(["Tracing?", "Yes" if self.trace is not None else "No"])

        """
        for reg in self._info['context']:
            table.add_row([reg, hex(getattr(self, reg))])
        
        """
        table.header = False
        table.align = "l"

        return str(table) + '\n' + str(self.context)


    @property
    def id(self) -> int:
        """Thread ID"""
        return self._info['id']

    @property
    def state(self) -> str:
        """Thread state, such as 'waiting', 'suspended'"""
        return self._info['state']

    @property
    def pc(self) -> int:
        """The current program counter/instruction pointer."""
        return int(self._info['context']['pc'],16)

    @property
    def module(self) -> str:
        """What module is the thread's program counter in? i.e.:
        libc-2.27.so."""
        return self._process.get_module_by_addr(self.pc) or "Unknown"
    
    @property
    def trace(self):
        """revenge.tracer.instruction_tracer.Trace: Returns Trace object if this thread is currently being traced, otherwise None."""
        if self.id in self._process.tracer._active_instruction_traces:
            return self._process.tracer._active_instruction_traces[self.id]


class Threads(object):

    def __init__(self, process):
        self._process = process

    def __len__(self):
        return len(self.threads)

    def __iter__(self):
        return iter(self.threads)

    def __repr__(self):
        return "<{} {}>".format(len(self), "Thread" if len(self) == 1 else "Threads")

    def __str__(self):
        table = PrettyTable(['id', 'state', 'pc', 'module', 'Trace'])

        for thread in self:
            table.add_row([str(thread.id), thread.state,
                self._process.memory.describe_address(thread.pc).split(":")[-1], thread.module, 'Yes' if thread.trace is not None else 'No'])

        return str(table)

    def __getitem__(self, elm):

        if type(elm) is int:
            try:
                return next(thread for thread in self.threads if thread.id == elm)
            except StopIteration:
                logger.error("Invalid thread id selected.")

        else:
            logger.error("Not sure how to handle this.")

    @property
    def threads(self) -> typing.List[Thread]:
        """Current snapshop of active threads."""
        threads = self._process.run_script_generic("""send(Process.enumerateThreadsSync());""", raw=True, unload=True)[0][0]
        return [Thread(self._process, thread) for thread in threads]

from .tracer.contexts import Context




import logging
logger = logging.getLogger(__name__)

import collections

class Java:

    def __init__(self, process):
        """Handles performing Java related activities."""
        self._process = process
        
        # Key = Full path to method that was implemented, Value = str that we implemented with.
        self._implementations = {}

        # Key: class name, vlaue = list of handles to active objects in memory
        self._active_handles = collections.defaultdict(lambda: list())

        # Key: Full class name, value = list of dict of reflected info
        self._cache_reflected_methods = collections.defaultdict(lambda: list())
        self._cache_reflected_fields = collections.defaultdict(lambda: list())
    
    def run_script_generic(self, script_name, raw=False, main_thread=False, *args, **kwargs):
        """Run the given Java related Frida calls. Simply wraps them in the perform call...

        Java Specific:
        Args:
            main_thread (bool, optional): Run this on the main Java thread.

        Calls the Process.run_script_generic as below:
        """

        context = kwargs.get("context", None)

        if not raw:
            script = self._process.load_js(script_name)
        else:
            # NOTE: This is meant to transparently convert the java_class and
            # others into the corresponding code. Do not remove str call!
            script = str(script_name)

        # If we're outside a context, do the full setup
        if context is None:
            action = "Java.perform( function () {"
            action_end = "});"

            if main_thread:
                script = "Java.scheduleOnMainThread( function () { " + script + "});"

            # Wrap up the java call
            script = "Java.perform( function () {" + script + "});"

        #kwargs['timeout'] = None

        return self._process.run_script_generic(script, raw=True, *args, **kwargs)

    def find_active_instance(self, klass, invalidate_cache=False):
        """Look through memory and finds an active instance of the given klass.

        Args:
            klass (str, JavaClass): The class we want to find already in memory.
            invalidate_cache (bool, optional): Throw away any current cache.
                This should normally not be needed.

        Returns:
            Returns JavaClass instance with approrpiate handle server. This
            means you can use the object without instantiating it yourself.

        Example:
            MainActivity = p.java.find_active_class("ooo.defcon2019.quals.veryandroidoso.MainActivity")
            MainActivity.parse("test")
        """

        if isinstance(klass, JavaClass):
            klass_name = klass._name
            
        elif isinstance(klass, str):
            klass_name = klass

        else:
            logger.error("Invalid klass type of {}".format(type(klass)))
            return

        logger.warning("There's currently a bug in this from frida server version 12.6.12 - current. If this fails to work for you, install server version 12.6.11 via android = revenge.device_types.AndroidDevice(type='usb', frida_server_release='12.6.11')")

        if invalidate_cache:
            self._active_handles[klass_name] = []

        # Attempt to enumerate active handles if we don't know of any.
        if self._active_handles[klass_name] == []:
            self._active_handles[klass_name] = self.run_script_generic("var my_list = []; Java.choose('{}', {{onMatch: function (i) {{my_list.push(i); send(i);}}, onComplete: function () {{send('DONE');}}}})".format(klass_name), raw=True, unload=False,onComplete='DONE')[0]

        # If we can't find any instances
        # TODO: Clean-up script if we don't find anything..
        if self._active_handles[klass_name] == []:
            logger.warn("Couldn't find any active instances of {}!".format(klass_name))
            return

        # Build new instance
        handle = common.auto_int(self._active_handles[klass_name][0]['$handle'])
        klass = JavaClass(self._process, klass_name, handle=handle)
        return klass

    @property
    def classes(self):
        """JavaClasses: Returns java classes object."""
        return JavaClasses(self._process)

    @property
    def BatchContext(self):
        """Returns a BatchContext class for this jvm.

        Example:
            with process.java.BatchContext() as context:
                <stuff>

        BatchContext doc:
        """
        return lambda **kwargs: BatchContext(self._process,
                run_script_generic=self.run_script_generic, 
                handler_pre = "Java.perform(function () {",
                handler_post = "});",
                **kwargs)

from .classes import JavaClasses
from .java_class import JavaClass
from ..process import Process
from ..contexts.batch import BatchContext
from .. import common

# Fixup docs
Java.run_script_generic.__doc__ += Process.run_script_generic.__doc__
Java.BatchContext.__doc__ += BatchContext.__init__.__doc__




import logging
logger = logging.getLogger(__name__)

import multiprocessing
import time
import json
import collections
from termcolor import cprint, colored

from prettytable import PrettyTable

from .. import types, common
from ..threads import Thread

class TraceItem(object):

    def __init__(self, process, item):
        self._process = process
        self._item = item
        self.from_ip = None
        self.from_module = None
        self.to_ip = None
        self.to_module = None
        self.type = None
        self.depth = None

        self._parse_item(item)

    def _parse_item(self, item):

        # Common
        self.type = item['type']
        self.from_ip = types.Pointer(common.auto_int(item['from_ip']))
        self.from_module = item['from_module']

        if 'to_ip' in item:
            self.to_ip = types.Pointer(common.auto_int(item['to_ip']))

        if 'to_module' in item:
            self.to_module = item['to_module']

        if 'depth' in item:
            self.depth = common.auto_int(item['depth'])

    def _str_add_table_row(self, table):
        
        if self.depth is not None:
            indent = ' '*self.depth
        else:
            indent = ''

        table.add_row([
            colored(self.type, attrs=['bold']),
            indent + self._process.memory.describe_address(self.from_ip, color=True),
            indent + self._process.memory.describe_address(self.to_ip, color=True) if self.to_ip is not None else "",
            str(self.depth) if self.depth is not None else ""
            ])

    def __str__(self):

        table = PrettyTable(['Type', 'From', 'To', 'Depth'])
        table.border = False
        table.header = False

        self._str_add_table_row(table)

        return str(table)

    def __repr__(self):
        attrs = ["TraceItem"]
        attrs.append(hex(self.from_ip))
        attrs.append(self.type)

        return "<{}>".format(' '.join(attrs))

    @property
    def type(self):
        return self.__type

    @type.setter
    def type(self, t):
        assert isinstance(t, (str, type(None))), "Invalid type for type of {}".format(type(t))

        if t is None:
            self.__type = None
            return

        t = t.lower()

        if t not in ['call', 'ret', 'exec', 'block', 'compile']:
            logger.error("Unhandled traceitem type of {}".format(t))
            logger.error(str(self._item))
            return

        self.__type = t



class Trace(object):
    """Keeps information about a Trace."""
    
    def __init__(self, process, tid, script, callback=None):
        self._process = process
        self._trace = []
        self._tid = tid
        self._script = script
        self._callback = callback

    def append(self, item):
        ti = TraceItem(self._process, item)
        self._trace.append(ti)

        if self._callback is not None:
            self._callback(self._tid, ti)

    def stop(self):
        """Stop tracing."""

        if self._script is not None:
            # TODO: Why the hell is Frida freezing on attempting to unload the stalker script?
            # Must unfollow a Stalked thread in the SAME CONTEXT IT IS STALKING! Thus the RPC export here.
            self._script[0].exports.unfollow()
            # TODO: Add unload back in once it doesn't take forever for it to unload the script...
            # Until then, calling unfollow and not unloading the script seems to be OK.
            #time.sleep(1)
            #self._script[0].unload()
            self._process.tracer._active_instruction_traces.pop(self._tid)
            self._script = None

    def wait_for(self, address):
        """Don't return until the given address is hit in the trace."""
        address = self._process._resolve_location_string(address)

        # TODO: Optimize this so I don't keep checking the same IPs over and over
        while True:
            try:
                next(x for x in self._trace if x.from_ip == address)
                break
            except StopIteration:
                continue
        
    def __iter__(self):
        return (x for x in self._trace)

    def __len__(self):
        return len(self._trace)

    def __str__(self):
        table = PrettyTable(['Type', 'From', 'To', 'Depth'])
        table.border = False
        table.header = False
        table.align = 'l'

        depth = 0

        for i in self:
            # Implicitly assign depths
            if i.depth is None:
                i.depth = depth
            
            i._str_add_table_row(table)

            if i.type == 'call':
                depth = i.depth + 1
            elif i.type == 'ret':
                depth = i.depth - 1

        return str(table)

    def __repr__(self):
        attr = ['Trace', 'Thread={}'.format(self._tid)]
        attr += [str(len(self)), 'items']

        return "<{}>".format(' '.join(attr))

    def __getitem__(self, item):

        if isinstance(item, int):
            return self._trace.__getitem__(item)

        if isinstance(item, slice):
            ret = Trace(self._process, self._tid, script=None)
            ret._trace = self._trace[item]
            return ret

        raise Exception("Unhandled getitem type of {}".format(type(item)))

class InstructionTracer(object):

    def __init__(self, process, threads=None, from_modules=None, call=False,
            ret=False, exec=False, block=False, compile=False, callback=None):
        """

        Args:
            process: Base process instantiation
            threads (list, optional): What threads to trace. If None, it will trace all threads.
            from_modules (list, optional): Restrict trace returns to those that start from one of the listed modules.
            call (bool, optional): Trace calls
            ret (bool, optional): Trace rets
            exec (bool, optional): Trace all instructions
            block (bool, optional): Trace blocks
            compile (bool, optional): Trace on Frida instruction compile
            callback (callable, optional): Callable to call with list of new
                instructions as they come in. First arg will be the thread id.
        """

        assert callable(callback) or callback is None, "Invalid type for callback of {}".format(type(callback))

        # Santiy check
        if not any((call, ret, exec, block, compile)):
            error = "You didn't select any action to trace!"
            logger.error(error)
            #raise Exception(error)

        self._process = process
        self.call= call
        self.ret = ret
        self.exec = exec
        self.block = block
        self.compile = compile
        self.threads = threads
        self._script = {}
        self._from_modules = from_modules
        self.callback = callback

        # IMPORTANT: It's important to keep a local pointer to this trace. It's
        # possible for trace messages to come in after officially stopping the
        # trace. Using local dict in this way allows this trace to continue to
        # get information while still being stopped.
        self.traces = {}

        self._start()

    def _on_message(self, m, d):
        try:
            payload = m['payload']
        except:
            print(m)
            raise

        for x in payload:
            self.traces[x['tid']].append(x)

    def _start(self):

        replace = {
            "FROM_MODULES_HERE": json.dumps([module.name for module in self._from_modules]),
            "STALK_CALL": json.dumps(self.call),
            "STALK_RET": json.dumps(self.ret),
            "STALK_EXEC": json.dumps(self.exec),
            "STALK_BLOCK": json.dumps(self.block),
            "STALK_COMPILE": json.dumps(self.compile),
        }

        for thread in self.threads:
            replace['THREAD_ID_HERE'] = str(thread.id)
            self._process.run_script_generic("stalk.js", replace=replace, unload=False, on_message=self._on_message, runtime='v8')
            self.traces[thread.id] = Trace(self._process, thread.id, self._process._scripts.pop(0), callback=self.callback)
            self._process.tracer._active_instruction_traces[thread.id] = self.traces[thread.id]

    def __repr__(self):
        attrs = ["InstructionTracer"]
        attrs += [str(len(self.threads)), 'threads']

        return "<{}>".format(' '.join(attrs))

    def __iter__(self):
        return self.traces.values().__iter__()

    def __str__(self):
        table = PrettyTable(['tid', 'count'])

        for tid, trace in self.traces.items():
            table.add_row([str(tid), str(len(trace))])

        return str(table)

    @property
    def threads(self):
        """list: Threads that are being traced by this object."""
        return self.__threads

    @threads.setter
    def threads(self, threads):
        assert isinstance(threads, (type(None), list, tuple, Thread)), "Invalid threads type of {}".format(type(threads))

        if threads is None:
            threads = list(self._process.threads)

        if not isinstance(threads, (list, tuple)):
            threads = [threads]

        else:
            threads_new = []
            for thread in threads:
                if isinstance(thread, Thread):
                    threads_new.append(thread)
                elif isinstance(thread, int):
                    threads_new.append(self._process.threads[thread])
                else:
                    raise Exception("Unable to resolve requested thread of type {}".format(type(thread)))

            threads = threads_new

        # Make sure the threads aren't already being traced
        for thread in threads:
            if thread.id in self._process.tracer._active_instruction_traces:
                error = "Cannot have more than one trace on the same thread at a time. Stop the existing trace with: process.threads[{}].trace.stop()".format(thread.id)
                logger.error(error)
                raise Exception(error)

        self.__threads = threads

    @property
    def _from_modules(self):
        """list,tuple,str,Module,None: What modules to restrict tracing from. Items can be strings (which will resolve) or Module objects."""
        return self.__from_modules

    @_from_modules.setter
    def _from_modules(self, modules):

        assert isinstance(modules, (list, tuple, type(None), str, Module)), "Unsupported type for from_modules of {}".format(type(modules))
        
        if modules is None:
            self.__from_modules = []
            return
        
        if not isinstance(modules, (list, tuple)):
            modules = [modules]

        new_modules = []
        for module in modules:
            if isinstance(module, Module):
                new_modules.append(module)
            elif isinstance(module, str):
                new_modules.append(self._process.modules[module])
            else:
                error = "Unsupported type for module of {}".format(type(module))
                logger.error(error)
                raise Exception(error)
        
        self.__from_modules = new_modules

from ..modules import Module




import logging
logger = logging.getLogger(__name__)

from elftools.elf.elffile import ELFFile
from termcolor import cprint, colored
import itertools
import hashlib
import os
import io
import json
from fnmatch import fnmatch
import pefile

from .. import common, types, config, symbols

symbol_cache_path = os.path.join(config.app_dirs.user_cache_dir, 'symbol_cache')
os.makedirs(symbol_cache_path, exist_ok=True)

class Module(object):

    def __init__(self, process, name, base, size, path):
        self._process = process
        self.name = name
        self.base = base
        self.size = size

        self.path = path # Must go last

    def _read_symbols_cache_dict(self, f):
        """f == file like object

        Returns either None (if no cache exists) or dict of cache."""

        if f is None:
            return None

        assert isinstance(f, io.IOBase), 'Unhandled symbol cache load type of {}'.format(type(f))

        f.seek(0, 0)
        h = hashlib.sha256(f.read()).hexdigest()
        f.seek(0, 0)

        this_cache_file = os.path.join(symbol_cache_path, h)

        # Cache miss
        if not os.path.isfile(this_cache_file):
            return None

        # Cache hit
        with open(this_cache_file, "r") as f:
            return json.loads(f.read())

    def _load_symbols_cache(self, cache):
        """Loads symbols previously discovered.

        Args:
            cache (dict): Dictionary cache to load up.
        """

        #
        # Symbols
        #

        for sym, address in cache['symbols'].items():
            if self._process.file_type is "PE" or \
                    (self.elf is not None and self.elf.type_str == 'DYN'):
                address = address + self.base

            sym = symbols.Symbol(self._process, name=sym, address=address)
            self._process.modules._symbol_to_address[self.name][sym.name] = sym
            self._process.modules._address_to_symbol[address] = sym

        #
        # Sections
        #

        if 'plt_offset' in cache:
            self.plt = cache['plt_offset']

        #
        # PLT
        #

        if 'plt' in cache:
            for i, sym in enumerate(cache['plt']):
                addr = self.plt + ((i+1)*0x10)
                name = 'plt.' + sym
                sym = symbols.Symbol(self._process, name=name, address=addr)
                self._process.modules._symbol_to_address[self.name][name] = sym
                self._process.modules._address_to_symbol[addr] = sym

    def _save_symbols_cache(self, file_io, cache):
        """Saves symbols into cache to be used later.

        Args:
            file_io (file like): The base file in full.
            cache (dict): The symbols we discovered.
        """

        file_io.seek(0, 0)
        h = hashlib.sha256(file_io.read()).hexdigest()
        file_io.seek(0, 0)

        this_cache_file = os.path.join(symbol_cache_path, h)
        with open(this_cache_file, "w") as f:
            f.write(json.dumps(cache))
        

    def _load_symbols(self):
        """Reads in the file for this module and attempts to extract the symbols."""

        # Either we're loading everything or what we're looking at right now __should__ be loaded
        if self._process._load_symbols is None or any(True for x in self._process._load_symbols if fnmatch(self.name, x)):

            # Clear out old symbols if needed
            self._process.modules._symbol_to_address[self.name] = {}

            file_io = common.load_file(self._process, self.path)
            cache = self._read_symbols_cache_dict(file_io)

            if cache is not None:
                return self._load_symbols_cache(cache)
        
            if self._process.file_type == 'ELF':
                self._load_symbols_elf(file_io)
            
            elif self._process.file_type == "PE":
                self._load_symbols_pe(file_io)

            # TODO: Windows
            # TODO: Mac

        # If we didn't resolve anything, make sure we noted we tried
        if self.name not in self._process.modules._symbol_to_address:
            self._process.modules._symbol_to_address[self.name] = {}

    def _load_symbols_pe(self, pe_io):
        # TODO: Assuming that this process will work on any system running ELF...
        print("Loading symbols for {} ... ".format(self.name), end='', flush=True)

        pe = pefile.PE(data=pe_io.read())
        cache = {'symbols': {}}

        # Some PEs don't export anything.
        if hasattr(pe, "DIRECTORY_ENTRY_EXPORT"):
            for sym in pe.DIRECTORY_ENTRY_EXPORT.symbols:
                if sym.name in [b'', None]:
                    continue

                name = sym.name.decode()

                rel_address = sym.address
                address = rel_address + self.base

                self._process.modules._symbol_to_address[self.name][name] = types.Pointer(address)
                self._process.modules._address_to_symbol[address] = name
                cache['symbols'][name] = rel_address

        self._save_symbols_cache(pe_io, cache)
        cprint("[ DONE ]", "green")

    def _load_symbols_elf(self, elf_io):
        # TODO: Assuming that this process will work on any system running ELF...
        print("Loading symbols for {} ... ".format(self.name), end='', flush=True)

        if elf_io is None:
            cprint("[ Failed to load ]", "yellow")
            return

        e = ELFFile(elf_io)

        #
        # Load up any symbols from the file
        #

        symtab = e.get_section_by_name('.symtab')
        dynsym = e.get_section_by_name('.dynsym')

        symbols = []
        cache = {
                'symbols': {},
                'plt': [],
                }

        #
        # Static Symbols
        #

        # Sometimes the binary won't have a symbol table
        if symtab is not None:
            symbols.append(symtab.iter_symbols())

        if dynsym is not None:
            symbols.append(dynsym.iter_symbols())

        # Pull out symbols
        for sym in itertools.chain(*symbols):
            if sym.name == '':
                continue

            address = sym['st_value']
            rel_address = address

            if self.elf.type_str == 'DYN':
                address = address + self.base

            cache['symbols'][sym.name] = rel_address

        #
        # Section offsets
        #

        plt_section = e.get_section_by_name('.plt')

        if plt_section is not None:

            cache['plt_offset'] = plt_section.header.sh_offset
            self.plt = cache['plt_offset']

            #
            # PLT/GOT
            #

            rels = []

            try:
                rels.append(e.get_section_by_name('.rela.plt').iter_relocations())
            except AttributeError:
                pass

            try:
                rels.append(e.get_section_by_name('.rel.plt').iter_relocations())
            except AttributeError:
                pass
            
            i = 1
            for s in itertools.chain(*rels):
                sym_name = dynsym.get_symbol(s.entry.r_info_sym).name
                got_offset = s.entry.r_offset

                if self.elf.type_str == 'DYN':
                    got = self.base + got_offset
                else:
                    got = got_offset

                cache['plt'].append(sym_name)
                cache['symbols']['got.' + sym_name] = got_offset
                i += 1

        else:
            logger.warning("Could not find PLT for ELF '{}'".format(self.name))

        self._save_symbols_cache(elf_io, cache)
        self._load_symbols_cache(cache)

        cprint("[ DONE ]", "green")

    def __repr__(self):
        attrs = ['Module', self.name, '@', hex(self.base)]
        return "<{}>".format(' '.join(attrs))

    def __eq__(self, other):
        return self.name == other.name and self.base == other.base and self.path == other.path and self.size == other.size

    @property
    def name(self):
        """str: Module name."""
        return self.__name

    @name.setter
    def name(self, name):
        if type(name) is not str:
            error = "Name must be string, not {}".format(type(name))
            logger.error(error)
            raise Exception(error)

        self.__name = name

    @property
    def base(self):
        """int: Base address this module is loaded at."""
        return self.__base

    @base.setter
    def base(self, base):

        base = common.auto_int(base)

        if type(base) is int:
            base = types.Pointer(base)

        self.__base = base

    @property
    def path(self):
        """str: Module path."""
        return self.__path

    @path.setter
    def path(self, path):
        if type(path) is not str:
            error = "Path must be string, not {}".format(type(path))
            logger.error(error)
            raise Exception(error)

        self.__path = path

        # Load up the symbols for this file if we haven't already
        if self.name not in self._process.modules._symbol_to_address:
            self._load_symbols()
        
    @property
    def size(self):
        """int: Size of this module."""
        return self.__size

    @size.setter
    def size(self, size):
        self.__size = common.auto_int(size)

    @property
    def elf(self):
        """Returns ELF object, if applicable, otherwise None."""

        if self._process.file_type != 'ELF':
            return

        try:
            return self.__elf
        except AttributeError:
            self.__elf = ELF(self._process, self)
            return self.__elf

    @property
    def symbols(self):
        """dict: symbol name -> address for this binary."""
        return self._process.modules._symbol_to_address[self.name]

    @property
    def plt(self):
        """int: Location of PLT for this module. Returns None if not known."""
        try:
            return self.symbols['.plt']
        except KeyError:
            return None

    @plt.setter
    def plt(self, address):

        if self._process.file_type == "PE":
            logger.warn("PLT is not valid for this file type.")
            return

        # Assuming plt needs to be rebased
        #if self.elf is not None and self.elf.type_str == 'DYN':
        address = address + self.base

        self.symbols['.plt'] = address


from ..parsers import ELF




import logging
logger = logging.getLogger(__name__)

import frida
import shlex
import os
import subprocess
from time import sleep

from .. import BaseDevice

class AndroidDevice(BaseDevice):
    def __init__(self, id=None, type=None, frida_server_release=None):
        """Describes and Android device.

        Args:
            id (str, optional): Unique ID for this android device
            type (str, optional): Connection type for this device.
            frida_server_release (str, optional): Specify a specific version
                i.e.: 12.6.11 -- default is latest

        Examples:
            AndroidDevice(id="emulator-5554")
            AndroidDevice(type="usb")
        """
        self.id = id
        self._frida_server_release = frida_server_release
        self.type = type.lower()
        self._select_device()

        # Make sure we can have root
        self.adb("root")

        self.start_frida_server()
        self.version # Load up version

    def _select_device(self):
        """Figure out which device we want to use."""

        if self.id:
            self.device = frida.get_device(self.id)
            return

        if self.type == "usb":
            usbs = [x for x in frida.enumerate_devices() if x.type == 'usb']
            if usbs == []:
                error = "Cannot find USB device."
                logger.error(error)
                raise Exception(error)

            if len(usbs) > 1:
                logger.warn("Found more than 1 usb device. Selecting first one...")

            self.device = usbs[0]
            self.id = self.device.id
            return

        error = "Couldn't find the device you requested..."
        logger.error(error)
        raise Exception(error)

    def _wait_for_frida_server(self):
        """To help avoid race conditions, pause here until the frida server is up and running."""
        while not self.frida_server_running:
            sleep(0.1)

    def start_frida_server(self):
        if self.frida_server_running:
            logger.info("Frida server already running on this device.")
            return
        
        logger.info("Attempting to start up Frida server on device.")

        # Download the server
        server_bin = common.download_frida_server("android", self.arch, release=self._frida_server_release)

        print("Pushing frida server ... ", flush=True)
        self.adb("push " + server_bin + " /data/local/tmp/frida-server")

        print("Starting frida server ... ", flush=True)
        self.adb("shell chmod +x /data/local/tmp/frida-server")
        self.adb(["shell", "nohup /data/local/tmp/frida-server &>/dev/null &"])

        # Clean up after ourselves
        os.unlink(server_bin)

        # Wait for it to start
        self._select_device()
        self._wait_for_frida_server()

    def adb(self, command, interactive=False):
        """Helper wrapper to run the command on your instance."""

        if isinstance(command, str):
            command = shlex.split(command)

        if not interactive:
            return subprocess.check_output(["adb", "-s", self.device.id] + command)
        return subprocess.call(["adb", "-s", self.device.id] + command)

    def shell(self):
        """Spawn and interact with a shell on the device."""
        self.adb("shell", interactive=True)


    def spawn(self, application, gated=True, load_symbols=None):
        """Spawn the given application.

        Args:
            application (str): Full application name (i.e.: com.android.calculator2)
            gated (bool, optional): If True, pause application immediately on loading
                This allows hooking prior to program startup.
            load_symbols (list): Only load symbols for the given modules. Same
                usage as revenge.Process

        Returns:
            revenge.Process: Process instantitation for this new process.
        """
        
        if gated:
            self.device.enable_spawn_gating()
        else:
            self.device.disable_spawn_gating()

        if not isinstance(application, frida._frida.Application):
            application = self.applications[application]

        application = application.identifier
        pid = self.device.spawn(application)

        if not gated:
            # Sometimes it gates anyway
            self.device.resume(pid)

        return Process(pid, device=self, load_symbols=load_symbols)

    def attach(self, application, load_symbols=None):
        """Attach to the given application.

        Args:
            application (str): Full application name (i.e.: com.android.calculator2) or pid
            load_symbols (list): Only load symbols for the given modules. Same
                usage as revenge.Process

        Returns:
            revenge.Process: Process instantitation for this new process.
        """
        
        if isinstance(application, frida._frida.Application):
            pid = application.pid
        else:
            pid = self.applications[application].pid

        return Process(pid, device=self, load_symbols=load_symbols)

    def install(self, package):
        """Install package onto android device.

        Args:
            package (str): Path to package to install.
        """

        package = os.path.abspath(package)

        if not os.path.isfile(package):
            logger.error("Cannot find apk to install.")
            return False

        # Allow replace by default
        return self.adb("install -r " + package)

    def uninstall(self, application):
        """Uninstall the given application.
        
        Args:
            application (str): Application to uninstall

        Example:
            android.uninstall("com.android.calculator2")
        """

        if isinstance(application, frida._frida.Application):
            application = application.identifier

        return self.adb("uninstall " + application)

    def __repr__(self):
        attrs = ['AndroidDevice', 'v' + self.version, self.id]
        return '<' + ' '.join(attrs) + '>'

    @property
    def applications(self):
        return AndroidApplications(self)

    @property
    def frida_server_running(self):
        try:
            self.device.enumerate_applications()
            return True
        except (frida.ServerNotRunningError, frida.InvalidOperationError, frida.ProcessNotFoundError, frida.TransportError):
            return False

    @property
    def arch(self):
        """Returns the arch of this android device."""
        try:
            return self.__arch
        except:
            #arch = subprocess.check_output(["adb", "-s", self.device.id, "shell", "uname", "-m"]).strip().decode()
            arch = self.adb("shell uname -m").strip().decode()

            if arch not in uname_standard:
                error = "Unhandled arch standardization of {}".format(arch)
                logger.error(error)
                raise Exception(error)

            self.__arch = uname_standard[arch]
            return self.__arch

    @property
    def version(self):
        """str: Returns android version for this device."""
        try:
            return self.__version
        except AttributeError:
            p = self.attach(self.device.get_frontmost_application(), load_symbols=[])
            self.__version = p.java.run_script_generic("send(Java.androidVersion)", raw=True, unload=True)[0][0]
            return self.__version

from .applications import AndroidApplications
from ... import common
from .. import uname_standard
from ...process import Process





import logging
logger = logging.getLogger(__name__)

#from .windows_messages import ActionWindowsMessages
from .stalker import ActionStalker
from .find import ActionFind
#from .differential_find import ActionDiffFind




import logging
logger = logging.getLogger(__name__)

import frida
import subprocess
import shlex
import os

# Standardize for Frida conventions
uname_standard = {
    "x86_64": "x86_64",
    "i686": "x86",
    "armv7l": "arm",
    "armv6l": "arm",
    "i386": "x86_64",
    "arm64": "arm64"
}

class BaseDevice:
    pass

class LocalDevice(BaseDevice):
    """Connect to whatever this is locally running on."""
    def __init__(self):
        self.device = frida.get_local_device()

from .android import AndroidDevice




import logging
logger = logging.getLogger(__name__)

from .. import common, types

class Tracer(object):

    def __init__(self, process):
        self._process = process

        # TID: Trace
        self._active_instruction_traces = {}

    def instructions(self, *args, **kwargs):
        """Start an instruction tracer."""
        return InstructionTracer(self._process, *args, **kwargs)

from . import InstructionTracer

# Fixup doc strings
Tracer.instructions.__doc__ = InstructionTracer.__init__.__doc__




import logging
logger = logging.getLogger(__name__)

from termcolor import cprint, colored

from ... import types, common

class SectionHeaders(object):

    def __init__(self, process, elf):
        self._process = process
        self._elf = elf
        self._headers = None

        # Give us ability to read
        mem = self._process.memory.maps[self.address]
        if mem.protection == '---':
            mem.protection = 'r--'
    
    def __len__(self):
        return self._elf.shnum

    def __iter__(self):
        return self._headers.__iter__()

    @property
    def _headers(self):
        """list: Headers for this elf."""
        if self.__headers is None:
            self._headers = []
            address = self.address
            for _ in range(len(self)):
                self._headers.append(SectionHeader(self._process, self._elf, address))
                address += self._elf.shentsize
        return self.__headers

    @_headers.setter
    def _headers(self, headers):
        self.__headers = headers

    @property
    def address(self):
        """ptr: Address where the section headers start."""
        return self._elf.shoff


from .section_header import SectionHeader




import logging
logger = logging.getLogger(__name__)

import colorama
from termcolor import cprint, colored
from ... import common
import json
import frida
import collections
from time import sleep

class ActionStalker:
    """General stalking action."""

    def __init__(self, process, include_module=None, tid=None, call=False, ret=False, exec=False, block=False, compile=False, include_function=None, *args, **kwargs):
        """
        Args:
            process: Parent process instantiation
            include_module: What module to follow, specifically
            tid (int, optional): What Thread ID to specifically follow (default: All)
        """

        self._process = process
        self._scripts = []
        self.from_modules = include_module or []
        self.tid = tid

        self.include_function = include_function
        if include_function is not None:
            logger.error("Sorry, include_function not working again just yet.")
            exit()
            function_module, function_offset = include_function.split(":")
            function_offset = int(function_offset, 16)
            self.include_function = self._process.modules[function_module]['base'] + function_offset
            logger.debug("Include function at: " + hex(self.include_function))

        self.call = call
        self.ret = ret
        self.exec = exec
        self.block = block
        self.compile = compile

        """
        # This should basically be moved over into the actual stalker lib

        self._include_function_traces = collections.defaultdict(lambda :list()) # key = pid, value = list/trace
        self._include_function_traces_depth = {}

        self.call = True if self.include_function is not None else call
        self.ret = True if self.include_function is not None else ret
        self.exec = exec
        self.block = True if self.include_function is not None else block
        self.compile = compile
        """

    def run(self):
        self.action_stalk()

    def action_stalk(self):
        """Start the stalker."""

        def stalk_cb(tid, ti):
            print(ti)
        
        trace = self._process.tracer.instructions(
                threads = self.tid,
                from_modules = self.from_modules,
                call = self.call,
                ret = self.ret,
                exec = self.exec,
                block = self.block,
                compile = self.compile,
                callback = stalk_cb,
                )

        # Make sure we're not blocking
        for addr in list(self._process.memory._active_breakpoints):
            self._process.memory[addr].breakpoint = False

        while self._process.alive:
            sleep(0.1)




import logging

logger = logging.getLogger(__name__)

import sys

if sys.version_info[0] < 3:
    logger.error('This script is supposed to be run with python3.')

from .process import Process
from . import symbols




import logging
logger = logging.getLogger(__name__)

import colorama
from termcolor import cprint, colored
from ... import common, types

import binascii
import time
import struct

class ActionFind:
    """Handle finding things in memory."""

    def __init__(self, process, include_module=None, string=None, uint8=None, 
            int8=None, uint16=None, int16=None, uint32=None, int32=None,
            uint64=None, int64=None, number=None, *args, **kwargs):
        """
        Args:
            process: Parent process instantiation
        """
        self._process = process
        self.include_module = include_module or []
        self.string = string
        self.uint8 = number if number else uint8
        self.int8 = number if number else int8
        self.uint16 = number if number else uint16
        self.int16 = number if number else int16
        self.uint32 = number if number else uint32
        self.int32 = number if number else int32
        self.uint64 = number if number else uint64
        self.int64 = number if number else int64
        self.number = number

        # Couple sanity checks
        if self._process.bits < 64:
            self.uint64 = None
            self.int64 = None

        if self._process.bits < 32:
            self.uint32 = None
            self.int32 = None

        self.discovered_locations = {}

    def run(self):
        # Sync
        self.action_find()

    def record_results(self, results, the_type):
        for r in results:
            self.discovered_locations[r] = the_type

    def action_find(self):

        if self.string is not None:
            r = self._process.memory.find(types.StringUTF16(self.string))
            r.sleep_until_completed()
            self.record_results(r, 'StringUTF16')

            r = self._process.memory.find(types.StringUTF8(self.string))
            r.sleep_until_completed()
            self.record_results(r, 'StringUTF8')

        if self.uint64 is not None:
            try:
                r = self._process.memory.find(types.UInt64(self.uint64))
                r.sleep_until_completed()
                self.record_results(r, 'UInt64')
            except struct.error:
                if self.number is None:
                    raise

        if self.int64 is not None:
            try:
                r = self._process.memory.find(types.Int64(self.int64))
                r.sleep_until_completed()
                self.record_results(r, 'Int64')
            except struct.error:
                if self.number is None:
                    raise

        if self.uint32 is not None:
            try:
                r = self._process.memory.find(types.UInt32(self.uint32))
                r.sleep_until_completed()
                self.record_results(r, 'UInt32')
            except struct.error:
                if self.number is None:
                    raise

        if self.int32 is not None:
            try:
                r = self._process.memory.find(types.Int32(self.int32))
                r.sleep_until_completed()
                self.record_results(r, 'Int32')
            except struct.error:
                if self.number is None:
                    raise

        if self.uint16 is not None:
            try:
                r = self._process.memory.find(types.UInt16(self.uint16))
                r.sleep_until_completed()
                self.record_results(r, 'UInt16')
            except struct.error:
                if self.number is None:
                    raise

        if self.int16 is not None:
            try:
                r = self._process.memory.find(types.Int16(self.int16))
                r.sleep_until_completed()
                self.record_results(r, 'Int16')
            except struct.error:
                if self.number is None:
                    raise

        # Ignoring exceptions when we're blanket searching for some 'number'
        if self.uint8 is not None:
            try:
                r = self._process.memory.find(types.UInt8(self.uint8))
                r.sleep_until_completed()
                self.record_results(r, 'UInt8')
            except struct.error:
                if self.number is None:
                    raise

        if self.int8 is not None:
            try:
                r = self._process.memory.find(types.Int8(self.int8))
                r.sleep_until_completed()
                self.record_results(r, 'Int8')
            except struct.error:
                if self.number is None:
                    raise





import logging
logger = logging.getLogger(__name__)

class Symbol(object):

    def __init__(self, process, name=None, address=None):
        """Represents a binary symbol.

        Args:
            process: Process object
            name (str, optional): Name of this symbol
            address (int, optional): Address of this symbol
        """

        self._process = process
        self.name = name
        self.address = address

    def startswith(self, x):
        return self.name.startswith(x)

    def __repr__(self):
        attrs = ['Symbol']

        if self.name is not None:
            attrs.append(self.name)
        
        if self.address is not None:
            attrs.append("@ " + hex(self.address))

        return "<" + ' '.join(attrs) + '>'

    def __str__(self):
        return self.name or ""

    def __int__(self):
        return self.address

    def __lt__(self, other):
        return self.address < other

    def __le__(self, other):
        return self.address <= other

    def __gt__(self, other):
        return self.address > other

    def __ge__(self, other):
        return self.address >= other

    def __sub__(self, other):
        return self.address - other

    def __add__(self, other):
        return self.address + other

    def __hex__(self):
        return hex(self.address)

    def __index__(self):
        return int(self.address)

    @property
    def name(self):
        """str: Name of this symbol."""
        return self.__name

    @name.setter
    def name(self, name):

        if not isinstance(name, str):
            logger.error("Symbol name must be string.")
            return

        self.__name = name

    @property
    def address(self):
        """int: Address of this Symbol."""
        return self.__address

    @address.setter
    def address(self, address):
        
        if not isinstance(address, int):
            logger.error("Symbol address must be an integer.")
            return

        self.__address = types.Pointer(address)

from . import types




import logging
logger = logging.getLogger(__name__)

from termcolor import cprint, colored

from ... import types, common

section_types = {
    0x0: 'SHT_NULL',
    0x1: 'SHT_PROGBITS',
    0x2: 'SHT_SYMTAB',
    0x3: 'SHT_STRTAB',
    0x4: 'SHT_RELA',
    0x5: 'SHT_HASH',
    0x6: 'SHT_DYNAMIC',
    0x7: 'SHT_NOTE',
    0x8: 'SHT_NOBITS',
    0x9: 'SHT_REL',
    0xa: 'SHT_SHLIB',
    0xb: 'SHT_DYNSYM',
    0xe: 'SHT_INIT_ARRAY',
    0xf: 'SHT_FINI_ARRAY',
    0x10: 'SHT_PREINIT_ARRAY',
    0x11: 'SHT_GROUP',
    0x12: 'SHT_SYMTAB_SHNDX',
    0x13: 'SHT_NUM',
}

class SectionHeader(object):

    def __init__(self, process, elf, address):
        self._process = process
        self._elf = elf
        self.address = address

    def __repr__(self):
        attrs = ['SectionHeader', self.type_str]
        return '<{}>'.format(' '.join(attrs))

    @property
    def name(self):
        """int: Index into shstrtab for the name of this section."""
        return self._process.memory[self.address].uint32

    @property
    def type(self):
        """int: Type of this section."""
        return self._process.memory[self.address+4].uint32

    @property
    def type_str(self):
        """str: Type of this section as str."""
        try:
            return section_types[self.type]
        except KeyError:
            return "Unknown"




import logging
logger = logging.getLogger(__name__)

class JavaClass(object):
    _in_init = set()

    def __init__(self, process, name=None, prefix=None, handle=None,
            full_description=None, is_method=None, is_field=None,
            implementation_message=None):
        """Represents a Java class.

        Args:
            process (revenge.Process): Process object
            name (str, optional): Full name for this class, method or field
            prefix (str, optional): What needs to be prefixed on to this object
                This is generally done automatically.
            handle (int, optional): Handle to where this class is instantiated
                in memory. Otherwise, it will be instantiated when used.
            full_description (str, optional): Automatically set full
                description of this method. Don't manually set.
            is_method (bool, optional): Is this a method
            is_field (bool, optional): Is this a field
            implementation_message (callable, optional): What should get called
                for any send that happen inside an implementation override?
                default: just print to the screen
        """

        JavaClass._in_init.add(id(self))

        self._reflection_done = False
        self._implementation_message = implementation_message
        self._is_method = is_method
        self._is_field = is_field
        self._class = None
        self._process = process
        self._name = name
        self._prefix = prefix or ""
        self._handle = handle
        self._full_description = full_description

        # Assume no prefix means this is the base class
        if self._prefix == "" and self._name != "":
            # class in front is needed to separate java class from built-in class
            self._class = "class " + self._name

            # Always reflect fields and methods on base class
            self._reflect_things()
        
        JavaClass._in_init.remove(id(self))

    def _reflect_things(self):

        if self._reflection_done:
            return

        # Methods shouldn't have reflect called on them.
        if self._is_method:
            return

        # Not Java class.
        if self._class is None or not self._class.startswith("class "):
            return

        self._reflect_fields()
        self._reflect_methods()

        self._reflection_done = True

    def _reflect_fields(self):
        """Reflectively identify fields."""

        assert self._class.startswith("class "), "Unexpected Class type {}".format(type(self._class))
        this_klass = self._class[6:]

        fields = self._process.java._cache_reflected_fields[this_klass]

        # If we missed the cache, enumerate it now
        if fields == []:
            fields = self._process.java.run_script_generic("get_declared_fields.js", unload=True, replace={'FULL_CLASS_HERE': this_klass})[0]

            # Save this off to the cache
            self._process.java._cache_reflected_fields[this_klass] = fields

        for field in fields:
            name = field['name']
            full_description = field['full_description']
            klass = field['class']

            # Just skipping unsafe names for now
            if not JavaClass._is_safe_method_name(name):
                continue

            # Create method instance
            # TODO: Change this to use Class init instead of setting manually
            new_class = JavaClass(self._process, name=name, prefix=str(self),
                    full_description=full_description, is_method=False,
                    is_field=True)
            new_class._class = klass
            setattr(self, name, new_class)
            
    def _reflect_methods(self):
        """Reflectively identify methods."""

        assert self._class.startswith("class "), "Unexpected Class type {}".format(type(self._class))
        this_klass = self._class[6:]

        methods = self._process.java._cache_reflected_methods[this_klass]

        # If we missed the cache, enumerate it now
        if methods == []:
            methods = self._process.java.run_script_generic("get_declared_methods.js", unload=True, replace={'FULL_CLASS_HERE': this_klass})[0]

            # Save this off to the cache
            self._process.java._cache_reflected_methods[this_klass] = methods

        for method in methods:
            name = method['name']
            full_description = method['full_description']

            # Just skipping unsafe names for now
            if not JavaClass._is_safe_method_name(name):
                continue

            # Create method instance
            # TODO: Change this to use Class init instead of setting manually
            new_class = JavaClass(self._process, name=name, prefix=str(self),
                    full_description=full_description, is_method=True,
                    is_field=False)
            setattr(self, name, new_class)

    def _parse_call_args(self, args):
        """Given args list, standardize it so it's ready for use in a call.
        
        Returns:
            list where each item is ready for inclusion.
        """
        ret_list = []

        for arg in args:
            if type(arg) is str:
                # Quotes and such
                ret_list.append(repr(arg))
            else:
                # This should work for int/float/etc as well as JavaClass
                ret_list.append(str(arg))

        return ret_list

    def __repr__(self):
        attrs = ["JavaClass"]
        
        if self._full_description is not None:
            attrs.append(self._full_description)

        elif self._name is not None:
            attrs.append(self._name)

        if self._handle is not None:
            attrs.append("Handle=" + str(self._handle))

        return "<" + " ".join(attrs) + ">"

    def __str__(self):

        # This is a direct class/method
        if self._name is not None:
            if not self._is_method and not self._is_field:

                ret = "Java.use('" + self._name + "')"

                # Are we using an already instantiated instance?
                if self._handle is not None:
                    ret = "Java.cast(ptr('{}'), ".format(hex(self._handle)) + ret + ")"

                return ret

            else:
                return self._prefix + "." + self._name
        else:
            # Name is none, just return prefix. We are probably drilled down.
            return self._prefix

    def __call__(self, *args, **kwargs):

        args = self._parse_call_args(args)

        # This is an actual call shorthand. We've made the line and want to run it.
        if self._prefix != "" and self._name is None:
            unload = kwargs.pop('unload', True)
            context = kwargs.pop('context', None)

            # Only send this back directly if we're not using a context
            if context is None:
                command = "send(" + str(self) + ")"
            else:
                command = str(self)

            ret = self._process.java.run_script_generic(command, raw=True, unload=unload, context=context, **kwargs)

            #
            # What to return from this
            #

            if ret is None or ret[0] == []:
                return None

            else:
                return ret[0][0]


        if not self._is_method:
            prefix = str(self) + ".$new(" + ",".join(args) + ")"
        else:
            prefix = str(self) + "(" + ",".join(args) + ")"

        return JavaClass(self._process, prefix=prefix)

    def __getattribute__(self, name):

        # Attribute miss will trigger getattr via exception
        ret = object.__getattribute__(self, name)

        # If this class is initializing, ignore this
        if id(self) in JavaClass._in_init:
            return ret
        
        # Hold off reflection until the last minute to avoid recurion and 
        # load costs
        if isinstance(ret, JavaClass):
            ret._reflect_things()

        return ret

    # Magic wrapper to drill down into methods
    def __getattr__(self, attr):
        return JavaClass(self._process, name=attr, prefix=str(self))

    @property
    def _is_method(self):
        """bool: Does this object actually describe a method?"""
        # If we've been explicitly told that this is a method
        if self.__is_method is not None:
            return self.__is_method

        # Infer that it is
        return self._prefix != "" and not self.__is_field

    @_is_method.setter
    def _is_method(self, is_method):
        assert isinstance(is_method, (bool, type(None))), "Invalid is_method type of {}".format(type(is_method))
        self.__is_method = is_method

    @property
    def _is_field(self):
        """bool: Is this a field?"""
        if self.__is_field is not None:
            return self.__is_field

        # Infer it
        return self._prefix != "" and not self.__is_method

    @_is_field.setter
    def _is_field(self, is_field):
        assert isinstance(is_field, (bool, type(None))), "Invalid is_field type of {}".format(type(is_field))
        self.__is_field = is_field

    @property
    def _is_class(self):
        """bool: Does this object actually describe a class?"""
        # If we've been explicitly told that this is a method
        if self.__is_class is not None:
            return self.__is_class

        # Infer that it is
        return self._prefix == ""

    @_is_class.setter
    def _is_class(self, is_class):
        assert isinstance(is_class, (bool, type(None))), "Invalid is_class type of {}".format(type(is_class))
        self.__is_class = is_class

    @property
    def _class(self):
        """str: The class type for this object."""
        return self.__class

    @_class.setter
    def _class(self, klass):
        assert isinstance(klass, (type(None), str)), "Invalid _class type of {}".format(type(klass))

        self.__class = klass

    @property
    def implementation(self):
        """str: Returns the over-written implementation for this method, or None if we have not over-written it.
        
        Examples:
            # Make random always return 5 (hey... it could be random?)
            process.java.classes['java.lang.Math'].random.implementation = "function (x) { return 5; }"
            assert process.java.classes['java.lang.Math'].random() == 5
            # Remove your overwrite so Random returns to normal
            process.java.classes['java.lang.Math'].random.implementation = None

            # Remember to call the actual function if you want to keep the functionality
            process.java.classes['java.lang.Math'].random.implementation = "function (x) { return this.random(x); }"
        """
        try:
            return self._process.java._implementations[str(self)]
        except KeyError:
            return None


    @implementation.setter
    def implementation(self, implementation):
        assert isinstance(implementation, (str, type(None))), "Unhandled implementation type of {}. Must be str type.".format(type(implementation))

        #
        # Regardless of what we're doing, we need to unimplement first.
        #

        try:
            script_stuff = self._process.java._implementations.pop(str(self))
            script_stuff[1].unload()

        except KeyError:
            # We didn't have an implementation yet. All good.
            pass

        # If we're only removing, then we're done here.
        if implementation is None:
            return

        #
        # Add new implementation
        #

        self._process.java.run_script_generic("{jclass}.implementation = {implementation}".format(
                jclass=str(self),
                implementation=implementation,
                ),
            raw=True,
            unload=False,
            runtime='v8',
            on_message=self._implementation_message,
            )
        
        # Save it off
        self._process.java._implementations[str(self)] = [implementation] + self._process._scripts.pop(0)

    @property
    def _implementation_message(self):
        """callable: Something to get called when 'send' messages are received from implementation."""
        return self.__implemenetation_message

    @_implementation_message.setter
    def _implementation_message(self, on_message):
        assert callable(on_message) or on_message is None, "Implementation message must be a callable!"

        # Default it to printing out the message
        if on_message is None:
            on_message = lambda x,y: common.on_msg_print(x,y,str(self))

        self.__implemenetation_message = on_message


    @staticmethod
    def _is_safe_method_name(name):
        safe = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_'
        return all(char in safe for char in name)


from .. import config
from .. import common




from . import Colorer
import logging

logger = logging.getLogger(__name__)

import typing
import frida
import colorama
colorama.init()

import os
from termcolor import cprint, colored
from prettytable import PrettyTable

here = os.path.dirname(os.path.abspath(__file__))

class NativeBacktrace(object):
    def __init__(self, process, backtrace):
        """Represents a backtrace. I.e.: what called what.

        Args:
            backtrace (list): List of instruction pointers
        """
        self._backtrace = backtrace

class NativeException(object):

    TYPES = ['abort', 'access-violation', 'illegal-instruction', 'arithmetic',
             'breakpoint', 'system']

    def __init__(self, context, backtrace=None, type=None,
            memory_operation=None, memory_address=None):
        """Represent a native CPU exception.
        
        Args:
            context: Frida-util cpu context
            backtrace: native backtrace object
            type (str): What type of exception is this.
            memory_operation (str, optional): Type of memory operation
                (read/write/execute)
            memory_address (int, optional): Address that was accessed when
                exception occurred.
        """

        self.context = context
        self.backtrace = backtrace
        self.type = type
        self.memory_operation = memory_operation
        self.memory_address = memory_address

    def __repr__(self):
        attrs = ['NativeException',
                self._process.memory.describe_address(self.address),
                self.type]

        return '<' + ' '.join(attrs) + '>'

    def __str__(self):
        s = "Native Exception\n"
        s += "~~~~~~~~~~~~~~~~\n"
        s += self.type + " at " + self._process.memory.describe_address(self.address) + "\n"

        if self.memory_operation is not None:
            s += "Memory " + self.memory_operation + " " + hex(self.memory_address) + "\n\n"
        else:
            s += "\n"

        s += str(self.context)
        s += "\n"

        # If we can't execute the memory location, don't print it
        if self.memory_operation != "execute":
            s += "\n" + str(self._process.memory[self.address].instruction_block)

        return s

    @classmethod
    def _from_frida_dict(cls, process, exception, backtrace):
        """Build a NativeException object directly from a frida dict."""
        assert isinstance(exception, dict)
        assert isinstance(backtrace, list)

        backtrace = NativeBacktrace(process, backtrace)
        return cls(
            context = CPUContext(process, **exception['context']),
            backtrace = backtrace,
            type = exception['type'],
            memory_operation = exception['memory']['operation'] if 'memory' in exception else None,
            memory_address = common.auto_int(exception['memory']['address']) if 'memory' in exception else None,
        )

    @property
    def _process(self):
        return self.context._process

    @property
    def type(self):
        """str: What type of native exception? One of """
        return self.__type

    @type.setter
    def type(self, type):
        type = type.lower()
        assert type in NativeException.TYPES, 'Unexpected native exception type of {}'.format(type)
        self.__type = type

    @property
    def address(self):
        """int: Address of this exception."""
        return self.context.ip

    @property
    def memory_address(self):
        """int: Address of memory exception."""
        return self.__memory_address

    @memory_address.setter
    def memory_address(self, memory_address):
        assert isinstance(memory_address, (int, type(None)))
        self.__memory_address = memory_address

    @property
    def memory_operation(self):
        """str: Type of memory operation performed at exception.

        Enum: read, write, execute"""
        return self.__memory_operation

    @memory_operation.setter
    def memory_operation(self, memory_operation):

        if isinstance(memory_operation, str):
            memory_operation = memory_operation.lower()
        assert memory_operation in ['read', 'write', 'execute', None], "Unexpected memory_operation of '{}'".format(memory_operation)
        
        self.__memory_operation = memory_operation

from .tracer.contexts import Context as CPUContext
from . import common

NativeException.type.__doc__ += ', '.join(NativeException.TYPES)



import logging
logger = logging.getLogger(__name__)

from .. import common, types

class MemoryMap(object):
    """Small wrapper to simply memory map lookups."""

    def __init__(self, process):
        self._process = process

        ranges = self._process.run_script_generic("""send(Process.enumerateRangesSync(''));""", raw=True, unload=True)[0][0]
        self._ranges = [MemoryRange(self._process, **range) for range in ranges]

    def __iter__(self):
        return self._ranges.__iter__()

    def __len__(self):
        return self._ranges.__len__()

    def __repr__(self):
        attrs = ['MemoryMap', str(len(self)), 'mapped ranges']
        return "<{}>".format(' '.join(attrs))

    def __str__(self):
        return str(self._process.memory)

    def __getitem__(self, item):

        if isinstance(item, int):
            for range in self:
                if item >= range.base and item <= range.base + range.size:
                    return range
            return None

        else:
            logger.error("Unsupported item type of {}".format(type(item)))

from . import MemoryRange




import logging
logger = logging.getLogger(__name__)

import os
import json
import io
import requests
import tempfile
import bs4
import lzma
import pprint

here = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(here, 'windows_messages_by_name.json')) as f:
    windows_messages_by_name = json.loads(f.read())

with open(os.path.join(here, 'windows_messages_by_id.json')) as f:
    windows_messages_by_id = json.loads(f.read())
    # JSON doesn't support int keys
    windows_messages_by_id = {int(x):y for x,y in windows_messages_by_id.items()}

with open(os.path.join(here, 'windows_keys_by_id.json')) as f:
    windows_keys_by_id = json.loads(f.read())
    # JSON doesn't support int keys
    windows_keys_by_id = {int(x):y for x,y in windows_keys_by_id.items()}

def auto_int(x):
    """Sometimes frida returns ints as a string instead of int. Just auto detect and return as int."""
    if isinstance(x, int):
        return x

    if isinstance(x, str):
        return int(x,0)

    # Accepting this as OK for now.
    if type(x) is float:
        return x

    logger.error("Unhandled auto_int type of {}".format(type(x)))

def parse_location_string(s):
    """Parse location string <[module:offset]|symbol> into (module, offset, symbol)."""

    assert type(s) == str, "Unexpected argument type of {}".format(type(s))

    module, offset_or_symbol = s.split(":")
    
    try:
        offset = hex(int(offset_or_symbol,0))
        symbol = ""
    except ValueError:
        offset = "0"
        symbol = offset_or_symbol

    return module, offset, symbol

def load_file(process, file_path):
    """Attempt to load the file with file_path. Use local loading if connection is local, and remote otherwise."""

    if process.device.device.type == 'local':
        return load_file_local(process, file_path)
    else:
        return load_file_remote(process, file_path)

def load_file_local(process, file_path):

    if not os.path.isfile(file_path):
        logger.debug("Couldn't load file: " + file_path)
        return

    return open(file_path, "rb")


def load_file_remote(process, file_path):
    """Attempt to load the file with file_path remotely, returning it in full as a BytesIO object."""

    if process.device_platform == 'linux':


        fopen = process.memory[':fopen']
        fseek = process.memory[':fseek']
        ftell = process.memory[':ftell']
        fread = process.memory[':fread']
        fclose = process.memory[':fclose']
        malloc = process.memory[':malloc']
        free = process.memory[':free']

        fp = fopen(file_path, 'r')

        # If we couldn't open it, fail gracefully
        if fp == 0:
            logger.debug("Couldn't load file: " + file_path)
            return

        fseek(fp, 0, 2)
        size = ftell(fp)
        fseek(fp, 0, 0)
        
        malloc_ptr = malloc(size)
        mem = process.memory[malloc_ptr:malloc_ptr+size]
        fread(malloc_ptr, size, 1, fp)

        elf_io = io.BytesIO(mem.bytes)
        free(malloc_ptr)
        fclose(fp)

        return elf_io

    else:
        logger.error("No remote file load support yet for: " + process.device_platform)

def download_frida_server(os, arch, release=None):
    """Download frida server of arch to a temporary file. Returns temporary file location.

    Args:
        os (str): What OS to download, i.e.: ios, android, linux, windows, macos
        arch (str): What arch to download, i.e.: x86, x86_64, arm, arm64
        release (str, optional): What release to download (default to latest)
            Example: 12.6.11
    
    Examples:
        download_frida_server('android', 'x86_64')
        download_frida_server('android', 'arm')
    """

    os = os.lower()
    arch = arch.lower()
    valid_os = ['android', 'ios', 'windows', 'macos']
    valid_arch = ['arm64', 'x86_64', 'x86', 'arm']

    assert os in valid_os, "Invalid OS selected. Must be in: " + str(valid_os)
    assert arch in valid_arch, "Invalid arch selected. Must be in: " + str(valid_arch)

    if release is None:
        download_url = "https://github.com/frida/frida/releases/latest"
    else:
        download_url = "https://github.com/frida/frida/releases/tag/" + release

    r = requests.get(download_url)
    html = bs4.BeautifulSoup(r.text, features="html.parser")
    download_links = set([x['href'] for x in html("a") if 'download' in x['href']])
    server_download_links = set([x for x in download_links if "frida-server" in x])

    look_for = "{os}-{arch}.".format(os=os, arch=arch)
    server_download_link = [x for x in server_download_links if look_for in x]

    if server_download_link == []:
        error = "Couldn't find a download link for {} {}!".format(os, arch)
        logger.error(error)
        raise Exception(error)

    if len(server_download_link) > 1:
        error = "Found multiple download links for {} {}!".format(os, arch)
        logger.error(error)
        raise Exception(error)

    server_download_link = "https://github.com" + server_download_link[0]
    print("Downloading " + server_download_link, flush=True)
    r = requests.get(server_download_link)
    server_bin = lzma.decompress(r.content)

    with tempfile.NamedTemporaryFile(delete=False) as fp:
        file_name = fp.name
        fp.write(server_bin)

    logger.debug("Server downloaded to: " + file_name)

    return file_name

def on_msg_print(m, d, TAG=None):
    """Generic on message handler to simply print it out."""

    if TAG is None:
        TAG = ''
    else:
        TAG = TAG + ": "

    if m['type'] == 'error':
        logger.error(TAG + "Script Run Error: " + pprint.pformat(m['description']))
        logger.debug(pprint.pformat(m))
        return
    
    print("on_message: {}".format([m,d]))




import logging
logger = logging.getLogger(__name__)

from fnmatch import fnmatch

class JavaClasses(object):

    def __init__(self, process):
        """Handles enumerating the loaded Java classes for this JVM."""
        self._process = process

    def __repr__(self):
        attrs = ["JavaClasses", str(len(self)), "classes"]
        return "<" + ' '.join(attrs) + ">"

    def __iter__(self):
        return self.classes.__iter__()

    def __len__(self):
        return len(self.classes)

    def __getitem__(self, item):
        if isinstance(item, int):
            return JavaClass(self._process, list(self)[item])

        elif isinstance(item, str):
            match = [JavaClass(self._process, x) for x in self if fnmatch(x, item)]
            if len(match) > 1:
                return match
            if match == []:
                return None
            return match[0]

        else:
            logger.error("Unhandled item get of type {}".format(type(item)))
            return

    @property
    def classes(self):
        """list: The actual list of classes."""
        try:
            return self.__classes
        except AttributeError:
            self.__classes = self._process.java.run_script_generic(r"""send( Java.enumerateLoadedClassesSync() );""", raw=True, unload=True)[0][0]
            return self.__classes

from .java_class import JavaClass



import logging
logger = logging.getLogger(__name__)

from termcolor import cprint, colored
from prettytable import PrettyTable

from ... import types, common

x64_regs = {
    'sp'  : 'self.esp',
    'bp'  : 'self.ebp',
    'ip'  : 'self.eip',
    'eax' : 'self.eax',
    'ax'  : 'self.eax & 0xffff',
    'al'  : 'self.eax & 0xff',
    'ah'  : '(self.eax>>8) & 0xff',
    'ebx' : 'self.ebx',
    'bx'  : 'self.ebx & 0xffff',
    'bl'  : 'self.ebx & 0xff',
    'bh'  : '(self.ebx>>8) & 0xff',
    'ecx' : 'self.ecx',
    'cx'  : 'self.ecx & 0xffff',
    'cl'  : 'self.ecx & 0xff',
    'ch'  : '(self.ecx>>8) & 0xff',
    'edx' : 'self.edx',
    'dx'  : 'self.edx & 0xffff',
    'dl'  : 'self.edx & 0xff',
    'dh'  : '(self.edx>>8) & 0xff',
    'esi' : 'self.esi',
    'si'  : 'self.esi & 0xffff',
    'sil' : 'self.esi & 0xff',
    'edi' : 'self.edi',
    'di'  : 'self.edi & 0xffff',
    'dil' : 'self.edi & 0xff',
    'ebp' : 'self.ebp',
    'bp'  : 'self.ebp & 0xffff',
    'bpl' : 'self.ebp & 0xff',
    'esp' : 'self.esp',
    'sp'  : 'self.esp & 0xffff',
    'spl' : 'self.esp & 0xff',
    'eip' : 'self.eip',
}

class X86Context(object):

    def __init__(self, process, **registers):
        """Represents a x86 CPU context.

        Example:
            X86Context(process, eax=12, ebx=13, <etc>)
        """

        self._process = process

        # Generically set any registers we're given
        for key, val in registers.items():
            setattr(self, key, common.auto_int(val))

    def __getattr__(self, attr):
        return eval(x64_regs[attr])

    def __str__(self):
        table = PrettyTable(["Register", "Value"])
        main_regs = ['eip', 'esp', 'ebp', 'eax', 'ebx', 'ecx', 'edx', 'esi', 'edi']

        table.align = 'l'

        for reg in main_regs:
            table.add_row([reg, hex(getattr(self, reg))])

        return str(table)




import logging
logger = logging.getLogger(__name__)

import colorama
from termcolor import cprint, colored
from .. import common
import json

class ActionWindowsMessages:
    """Handle stalking Windows Messages."""

    def __init__(self, process, include_module=None, windows_message=None, *args, **kwargs):
        """
        Args:
            process: Parent process instantiation
        """
        self._process = process
        self.include_module = include_module or []
        self.windows_message = windows_message
        self._scripts = []

    def run(self):
        self.action_windows_messages()

    def action_windows_messages(self):
        """Stalk some windows messages."""

        self._known_windows_message_handlers = []

        def windows_cb(message, data):
            # REMINDER: The JavaScript is filtering out dups. We will only be getting each handler once.

            handler_ip = int(message['payload'], 16)
            handler_module = self._process.get_module_by_addr(handler_ip)
            handler_offset = handler_ip - self._process.modules[handler_module]['base']

            self._known_windows_message_handlers.append(handler_ip)

            # Allow downselection to this module
            if self.include_module == [] or handler_module in self.include_module:
                print("{: <32}".format("Found Message Handler") + colored(handler_module, 'cyan') + ":" + colorama.Style.BRIGHT + colored(hex(handler_offset), "cyan"))
                self._action_windows_messages_intercept(handler_module, handler_offset)

        if self._process.device_platform != 'windows':
            logger.error('This doesn\'t appear to be a Windows device...')
            exit(1)

        windows_js = self._process.load_js('windows_stalk_message_handlers.js')

        script = self._process.session.create_script(windows_js)
        script.on('message', windows_cb)

        logger.debug("Starting Windows Message monitor ... ")
        script.load()

        # Save so that we don't GC it
        self._scripts.append(script)

    def _action_windows_messages_intercept(self, module, offset):
        """Start watching for windows events on this handler."""

        def window_message_cb(message, data):
            Hwnd, Msg, wParam, lParam, module, context, tid, depth = message['payload']
            # context is dict of regs. i.e.: context['pc'], context['rax'], etc
            tid = int(tid)
            depth = int(depth)
            Hwnd = int(Hwnd,16)
            Msg = int(Msg,16)
            wParam = int(wParam,16)
            lParam = int(lParam,16)

            try:
                msg = ','.join(colored(x,'magenta') for x in common.windows_messages_by_id[Msg])
            except KeyError:
                msg = 'Unknown (' + hex(Msg) + ')'

            extra = ""
            if Msg in common.windows_messages_by_id and any(True for x in common.windows_messages_by_id[Msg] if x in ['WM_SYSKEYDOWN', 'WM_SYSKEYUP', 'WM_KEYUP', 'WM_KEYDOWN', 'WM_CHAR']):
                key = common.windows_keys_by_id[wParam]
                extra = key['Constant'] + ": " + key['Description']

            else:
                extra = 'wParam: {} lParam: {}'.format(hex(wParam), hex(lParam))

            print('{module: <32}{msg} {extra}'.format(
                module=module,
                msg=msg,
                extra=extra,
                ))


        js = self._process.load_js('windows_intercept_message_handler.js')

        js = js.replace("OFFSET_HERE", hex(offset))
        js = js.replace("MODULE_HERE", module)

        constraints = ""

        # Constrain what messages we're looking at
        if self.windows_message is not None:
            constraints += 'if (! {}.includes(Number(args[1]))) {{ return; }}\n'.format(json.dumps(self.windows_message))

        js = js.replace('CONSTRAINTS_HERE', constraints)

        script = self._process.session.create_script(js, runtime='v8')
        script.on('message', window_message_cb)
        script.load()

        # Save so that we don't GC it
        self._scripts.append(script)





import logging
logger = logging.getLogger(__name__)

from .. import common, types

class MemoryRange(object):

    def __init__(self, process, base, size, protection, file=None):
        self._process = process
        self.base = base
        self.size = size
        self.protection = protection
        self._file = file

    def __repr__(self):
        value = ["MemoryRange", hex(self.base), '-', hex(self.base+self.size), self.protection]
        if self.file is not None:
            value.append(self.file + ":" + hex(self.file_offset))
        return '<' + ' '.join(value) + '>'


    @property
    def file(self):
        """str: File backing this memory range, or None."""
        if self._file is None:
            return None

        return self._file['path']

    @property
    def file_offset(self):
        """int: Offset into backing file or None."""
        if self._file is None:
            return None

        return self._file['offset']

    @property
    def readable(self):
        """bool: Is this range readable?"""
        return self.protection[0] == 'r'

    @property
    def writable(self):
        """bool: Is this range writable?"""
        return self.protection[1] == 'w'

    @property
    def executable(self):
        """bool: Is this range executable?"""
        return self.protection[2] == 'x'

    @property
    def protection(self):
        """str: Protection for this range."""
        return self.__protection

    @protection.setter
    def protection(self, protection):
        assert type(protection) is str
        assert len(protection) == 3
        protection = protection.lower()

        # If we're setting for the first time, assume it's correct
        if not hasattr(self, '_MemoryRange__protection'):
            self.__protection = protection

        # Set protection if it's not already this
        elif protection != self.protection:
            self._process.run_script_generic("""Memory.protect({}, {}, '{}')""".format(
                self.base.js,
                hex(self.size),
                protection,
                ), raw=True, unload=True)

            self.__protection = protection

    @property
    def size(self):
        """int: Size for this range."""
        return self.__size

    @size.setter
    def size(self, size):
        self.__size = common.auto_int(size)

    @property
    def base(self):
        """int: Base address for this range."""
        return self.__base

    @base.setter
    def base(self, base):
        self.__base = types.Pointer(common.auto_int(base))




import logging
logger = logging.getLogger(__name__)

from termcolor import cprint, colored

from ... import types, common

segment_types = {
    0: 'PT_NULL',
    1: 'PT_LOAD',
    2: 'PT_DYNAMIC',
    3: 'PT_INTERP',
    4: 'PT_NOTE',
    5: 'PT_SHLIB',
    6: 'PT_PHDR',
}

class ProgramHeader(object):

    def __init__(self, process, elf, address):
        self._process = process
        self._elf = elf
        self.address = address

    def __repr__(self):
        attrs = ['ProgramHeader', self.type_str]
        return '<{}>'.format(' '.join(attrs))

    @property
    def type(self):
        """int: What type of header is this."""
        return self._process.memory[self.address].uint32

    @property
    def type_str(self):
        """str: What type of header is this."""
        try:
            return segment_types[self.type]
        except KeyError:
            return "Unknown"

    @property
    def flags(self):
        """int: Flags for this header."""
        if self._elf.bits == 32:
            return self._process.memory[self.address+0x18].uint32
        else:
            return self._process.memory[self.address+0x4].uint32

    @property
    def vaddr(self):
        """int: Virtual address of the segment in memory."""
        if self._elf.bits == 32:
            return types.Pointer(self._elf.module.base + self._process.memory[self.address+0x8].pointer)
        else:
            return types.Pointer(self._elf.module.base + self._process.memory[self.address+0x10].pointer)

    @property
    def filesz(self):
        """int: Size in bytes of the segment in the file image."""
        if self._elf.bits == 32:
            return self._process.memory[self.address+0x10].uint32
        else:
            return self._process.memory[self.address+0x20].uint64

    @property
    def memsz(self):
        """int: Size in bytes of the segment in memory."""
        if self._elf.bits == 32:
            return self._process.memory[self.address+0x14].uint32
        else:
            return self._process.memory[self.address+0x28].uint64




import logging
logger = logging.getLogger(__name__)

# Keeping str types as properties in case they change what they call things later

class Basic: 
    def __add__(self, other):
        if type(self) is type(other) or type(other) is int:
            return self.__class__(int.__add__(self, other))
        else:
            logger.warning("Adding incompatible types {} and {}. Un-casting back to int.".format(type(self), type(other)))
            return int(self) + int(other)

    def __sub__(self, other):
        if type(self) is type(other) or type(other) is int:
            return self.__class__(int.__sub__(self, other))
        else:
            logger.warning("Subtracting incompatible types {} and {}. Un-casting back to int.".format(type(self), type(other)))
            return int(self) - int(other)

    @property
    def js(self):
        """String that can be fed into js."""
        return str(self)

class FloatBasic: 
    def __add__(self, other):
        if type(self) is type(other) or type(other) is int:
            return self.__class__(float.__add__(self, other))
        else:
            logger.warning("Adding incompatible types {} and {}. Un-casting back to float.".format(type(self), type(other)))
            return float(self) + float(other)

    def __sub__(self, other):
        if type(self) is type(other) or type(other) is int:
            return self.__class__(float.__sub__(self, other))
        else:
            logger.warning("Subtracting incompatible types {} and {}. Un-casting back to float.".format(type(self), type(other)))
            return float(self) - float(other)

class Int8(Basic, int):
    type = "int8"

class UInt8(Basic, int):
    type = "uint8"

class Int16(Basic, int):
    type = "int16"

class UInt16(Basic, int):
    type = "uint16"

class Int32(Basic, int):
    type = "int32"

class UInt32(Basic, int):
    type = "uint32"

class Int64(Basic, int):
    type = "int64"

    @property
    def js(self):
        return "int64('{}')".format(hex(self))

class UInt64(Basic, int):
    type = "uint64"

    @property
    def js(self):
        return "uint64('{}')".format(hex(self))

class Char(Int8):
    type = "char"

class UChar(UInt8):
    type = "uchar"

class Short(Int16):
    type = "short"

class UShort(UInt16):
    type = "ushort"

class Int(Int32):
    type = "int"

class UInt(UInt32):
    type = "uint"

class Long(Int64):
    type = "long"

class ULong(UInt64):
    type = "ulong"

class Float(FloatBasic, float):
    type = "float"

    @property
    def js(self):
        return str(self)

class Double(Float):
    type = "double"

class Pointer(UInt64):
    type = "pointer"

    @property
    def js(self):
        return "ptr('{}')".format(hex(self))

#
# These don't directly have a return type value, they will just be pointers..
# 

class StringUTF8(str):
    type = 'utf8'

    @property
    def js(self):
        logger.error("Shouldn't be asking for js on this object...")
        return str(self)

class StringUTF16(str):
    type = 'utf16'

    @property
    def js(self):
        logger.error("Shouldn't be asking for js on this object...")
        return str(self)


all_types = (Pointer, Int8, UInt8, Int16, UInt16, Int32, UInt32, Int64, UInt64, Char, UChar, Short, UShort, Int, UInt, Long, ULong, Float, Double, StringUTF8, StringUTF16)




import logging

logger = logging.getLogger(__name__)

from .x64 import X64Context
from .x86 import X86Context

def Context(process, *args, **kwargs):
    """Build context from args. Will auto discover context type.
    
    Example:
        context = Context(process, eax=1, ebx=2, ...)
    """

    arch = process.arch
    
    if arch == "x64":
        return X64Context(process, *args, **kwargs)

    elif arch == "ia32":
        return X86Context(process, *args, **kwargs)

    else:
        logger.error("Currently unsupported architecture of {}".format(arch))



import logging
logger = logging.getLogger(__name__)

from termcolor import cprint, colored
from prettytable import PrettyTable

from ... import types, common

x64_regs = {
    'sp'  : 'self.rsp',
    'bp'  : 'self.rbp',
    'ip'  : 'self.rip',
    'rax' : 'self.rax',
    'eax' : 'self.rax & 0xffffffff',
    'ax'  : 'self.rax & 0xffff',
    'al'  : 'self.rax & 0xff',
    'ah'  : '(self.rax>>8) & 0xff',
    'rbx' : 'self.rbx',
    'ebx' : 'self.rbx & 0xffffffff',
    'bx'  : 'self.rbx & 0xffff',
    'bl'  : 'self.rbx & 0xff',
    'bh'  : '(self.rbx>>8) & 0xff',
    'rcx' : 'self.rcx',
    'ecx' : 'self.rcx & 0xffffffff',
    'cx'  : 'self.rcx & 0xffff',
    'cl'  : 'self.rcx & 0xff',
    'ch'  : '(self.rcx>>8) & 0xff',
    'rdx' : 'self.rdx',
    'edx' : 'self.rdx & 0xffffffff',
    'dx'  : 'self.rdx & 0xffff',
    'dl'  : 'self.rdx & 0xff',
    'dh'  : '(self.rdx>>8) & 0xff',
    'r8'  : 'self.r8',
    'r8d' : 'self.r8 & 0xffffffff',
    'r8w' : 'self.r8 & 0xffff',
    'r8b' : 'self.r8 & 0xff',
    'r9'  : 'self.r9',
    'r9d' : 'self.r9 & 0xffffffff',
    'r9w' : 'self.r9 & 0xffff',
    'r9b' : 'self.r9 & 0xff',
    'r10' : 'self.r10',
    'r10d': 'self.r10 & 0xffffffff',
    'r10w': 'self.r10 & 0xffff',
    'r10b': 'self.r10 & 0xff',
    'r11' : 'self.r11',
    'r11d': 'self.r11 & 0xffffffff',
    'r11w': 'self.r11 & 0xffff',
    'r11b': 'self.r11 & 0xff',
    'r12' : 'self.r12',
    'r12d': 'self.r12 & 0xffffffff',
    'r12w': 'self.r12 & 0xffff',
    'r12b': 'self.r12 & 0xff',
    'r13' : 'self.r13',
    'r13d': 'self.r13 & 0xffffffff',
    'r13w': 'self.r13 & 0xffff',
    'r13b': 'self.r13 & 0xff',
    'r14' : 'self.r14',
    'r14d': 'self.r14 & 0xffffffff',
    'r14w': 'self.r14 & 0xffff',
    'r14b': 'self.r14 & 0xff',
    'r15' : 'self.r15',
    'r15d': 'self.r15 & 0xffffffff',
    'r15w': 'self.r15 & 0xffff',
    'r15b': 'self.r15 & 0xff',
    'rsi' : 'self.rsi',
    'esi' : 'self.rsi & 0xffffffff',
    'si'  : 'self.rsi & 0xffff',
    'sil' : 'self.rsi & 0xff',
    'rdi' : 'self.rdi',
    'edi' : 'self.rdi & 0xffffffff',
    'di'  : 'self.rdi & 0xffff',
    'dil' : 'self.rdi & 0xff',
    'rbp' : 'self.rbp',
    'ebp' : 'self.rbp & 0xffffffff',
    'bp'  : 'self.rbp & 0xffff',
    'bpl' : 'self.rbp & 0xff',
    'rsp' : 'self.rsp',
    'esp' : 'self.rsp & 0xffffffff',
    'sp'  : 'self.rsp & 0xffff',
    'spl' : 'self.rsp & 0xff',
    'rip' : 'self.rip',
}

class X64Context(object):

    def __init__(self, process, **registers):
        """Represents a x86_64 CPU context.

        Example:
            X64Context(process, rax=12, rbx=13, <etc>)
        """

        self._process = process

        # Generically set any registers we're given
        for key, val in registers.items():
            setattr(self, key, common.auto_int(val))

    def __getattr__(self, attr):
        return eval(x64_regs[attr])

    def __str__(self):
        table = PrettyTable(["Register", "Value"])
        main_regs = ['rip', 'rsp', 'rbp', 'rax', 'rbx', 'rcx', 'rdx', 'rsi', 'rdi', 'r8', 'r9', 'r10', 'r11', 'r12', 'r13', 'r14', 'r15']

        for reg in main_regs:
            table.add_row([reg, hex(getattr(self, reg))])

        return str(table)




import appdirs

def init():
    global app_dirs
    global recursion

    app_dirs = appdirs.AppDirs(appname='revenge', appauthor='bannsec')

    # Help watch for recursive loading
    recursion = set()

try:
    app_dirs
except NameError:
    init()




from . import Colorer
import logging
logging.basicConfig(level=logging.WARN)

logger = logging.getLogger(__name__)

import frida
import colorama
colorama.init()

import os
import sys
from termcolor import cprint, colored
from prettytable import PrettyTable
import time

import atexit
import signal
import json
import pprint
from copy import copy

here = os.path.dirname(os.path.abspath(__file__))

class Process(object):

    def __init__(self, target, resume=False, verbose=False, load_symbols=None,
            device=None):
        """

        Args:
            target (str, int): File name or pid to attach to.
            resume (bool, optional): Resume the binary if need be after loading?
            verbose (bool, optional): Enable verbose logging
            load_symbols (list, optional): Only load symbols from those modules
                in the list. Saves some startup time. Can use glob ('libc*')
            device (revenge.device_types.*, optional): Define what device
                to connect to.
        """

        # Just variable to ensure we don't garbage collect
        self._scripts = []
        # Cache common module addrs
        self._module_by_addr_cache = {}
        self.session = None
        self.__file_name = None
        self.__file_type = None
        self.__entrypoint = None
        self._resume_addr = None
        self.__endianness = None
        self.__bits = None
        self._spawn_target = None
        self.verbose = verbose
        self.device = device or device_types.LocalDevice()
        self.target = target

        if not isinstance(load_symbols, (list, type, type(None))):
            load_symbols = [load_symbols]
        self._load_symbols = load_symbols

        self.memory = Memory(self)
        self.threads = Threads(self)
        self.tracer = Tracer(self)
        self.modules = Modules(self)

        atexit.register(self._at_exit)
        self.start_session()

        if self.run_script_generic(r"""send(Java.available)""", raw=True, unload=True)[0][0]:
            self.java = Java(self)

        # ELF binaries start up in ptrace, which causes some issues, shim at entrypoint so we can remove ptrace
        if self._spawned_pid is not None and self.file_type == 'ELF':

            # Set breakpoint at entry
            self.memory[self.entrypoint].breakpoint = True

            # Set breakpoints at exit calls
            for c in [':exit', ':_exit']:
                self.memory[c].breakpoint = True

            # Resume to remove ptrace
            self.device.device.resume(self._spawned_pid)

            #time.sleep(0.2)

        if self.device_platform == 'linux':
            try:
                str(self.threads)
            except IndexError:
                logger.error("Can't enumerate threads. Please check sysctl kernel.yama.ptrace_scope=0 or run as root.")

        # Resume file if need be
        if resume:

            # If we are using a resume variable
            if self.memory[self.entrypoint].breakpoint:
                self.memory[self.entrypoint].breakpoint = False
            
            else:
                self.device.device.resume(self._spawned_pid)


    def pause_at(self, location):
        """Pause at a given point in execution."""

        pause_location = self._resolve_location_string(location)
        self.run_script_generic('pause_at2.js', replace={"FUNCTION_ADDRESS_HERE": hex(pause_location)})
        
    def _at_exit(self):
        """Called to clean-up at exit."""

        # TODO: Remove this once pytest fixes their at_exit issue
        logging.getLogger().setLevel(logging.WARN)

        try:

            # Remove breakpoints
            for addr in copy(self.memory._active_breakpoints):
                logger.debug("Removing breakpoint: " + hex(addr))
                self.memory[addr].breakpoint = False

            # Remove function replacements
            for addr in self.memory._active_replacements:
                self.memory[addr].replace = None

            # Remove any instruction traces
            for t in self.threads:
                if t.trace is not None:
                    t.trace.stop()

            self.run_script_generic("""Interceptor.detachAll()""", raw=True, unload=True)

            # Unallocate our memory
            for addr in copy(self.memory._allocated_memory):
                logger.debug("Unallocating memory: " + hex(addr))
                self.memory[addr].free()

            # Unload our scripts
            for script, text in self._scripts:
                logger.debug("Unloading Script: %s", text)

                try:
                    script.unload()
                except frida.InvalidOperationError:
                    # Already unloaded probably
                    pass

            logger.debug("Done unloading")

        except (frida.InvalidOperationError, frida.TransportError) as e:
            # Looks like the program terminated. Possibly finished execution before we finished cleanup.
            if 'detached' in e.args[0] or 'closed' in e.args[0]:
                return

        # If we spawned it, kill it
        try:
            if self._spawned_pid is not None:
                return self.device.device.kill(self._spawned_pid)

        except (frida.PermissionDeniedError, frida.ProcessNotFoundError) as e:
            # This can indicate the process is already dead.
            try:
                next(x for x in self.device.device.enumerate_processes() if x.pid == self._spawned_pid)
                logger.error("Device kill permission error, with process apparently %d still alive.", self._spawned_pid)
                raise e
            except StopIteration:
                return

        # Unload anything we loaded first
        #for script in self._scripts:
        #    script.unload()

        try:

            # Genericall unstalk everything
            for thread in self.threads:
                if thread.trace is not None:
                    thread.trace.stop()
                #self.run_script_generic("Stalker.unfollow({tid})".format(tid=thread.id), raw=True, unload=True)
        
        except frida.InvalidOperationError:
            # Session is already detached.
            pass

        # Detach our session
        self.session.detach()

    
    #######################
    # On Message Handlers #
    #######################

    def on_message(self, message, data=None):
        """Generic on message handler."""
        print("Caught message", message, data)


    def start_session(self):

        self._spawned_pid = None

        if self._spawn_target is not None:
            print("Spawning file\t\t\t... ", end='', flush=True)
            self._spawned_pid = self.device.device.spawn(self._spawn_target)
            cprint("[ DONE ]", "green")

        print('Attaching to the session\t... ', end='', flush=True)

        try:
            # Default attach to what we just spawned
            self.session = self.device.device.attach(self._spawned_pid or self.target)
        except frida.ProcessNotFoundError:
            logger.error('Could not find that target process to attach to!')
            exit(1)

        print(colorama.Fore.GREEN + '[ DONE ]' + colorama.Style.RESET_ALL)

    def target_type(self, x):
        # Maybe it's PID
        try:
            return int(x)

        # Probably process name
        except:
            return x

    def load_js(self, name):
        with open(os.path.join(here, "js", name), "r") as f:
            return f.read().strip()

    def get_module_by_addr(self, addr):

        addr = common.auto_int(addr)

        try:
            return self._module_by_addr_cache[addr]
        except:
            pass

        for module in self.modules:
            base = module.base
            size = module.size

            if addr >= base and addr <= base+size:
                self._module_by_addr_cache[addr] = module.name # Add to cache
                return module.name

        return None

    def run_script_generic(self, script_name, raw=False, replace=None,
            unload=False, runtime='duk', on_message=None, timeout=None,
            context=None, onComplete=None):
        """Run scripts that don't require anything special.
        
        Args:
            script_name (str): What script to load from the js directory
            raw (bool, optional): Should the script_name actually be considered
                the script contents?
            replace (dict, optional): Replace key strings from dictionary with
                value into script.
            unload (bool, optional): Auto unload the script. Set to true if the
                script is fully synchronous.
            runtime (str, optional): Runtime to use for this script, either
                'duk' or 'v8'.
            on_message(callable, optional): Set the on_message handler to this
                instead.
            timeout (int, optional): Modify timeout (default is 60 seconds).
                Note, this will cause the script to run async. 0 == no timeout
            context (Context, optional): Execute this script under a given
                context.
            onComplete (str, optional): If defined, this method will pause
                until the given onComplete string is returned. Basically
                allowing run_script_generic to return all things from an async
                script.

        Returns:
            tuple: msg, data return from the script
        """


        msg = []
        data = []

        # HACK: Using list for completed to make passing data back from async
        # function simpler.
        completed = [] # For use with async scripts

        if not on_message is None and not callable(on_message):
            logger.error('on_message handler must be callable.')
            return None

        def on_msg(m, d):

            if m['type'] == 'error':
                logger.error("Script Run Error: " + pprint.pformat(m['description']))
                logger.debug(pprint.pformat(m))
                return
            
            # Async is done
            if onComplete is not None and m['payload'] == onComplete:
                completed.append(True)
                return

            logger.debug("on_message: {}".format([m,d]))

            msg.append(m['payload'] if 'payload' in m else None)
            data.append(d)

        on_message = on_msg if on_message is None else on_message

        if not raw:
            js = self.load_js(script_name)
        else:
            js = script_name

        if replace is not None:
            assert type(replace) == dict, "Unexpected replace type of {}".format(type(replace))

            for key, value in replace.items():
                js = js.replace(key, value)

        if timeout is not None:
            js = "setTimeout(function() {" + js + "}," + str(timeout) + ")"
            # Forcing unload to false since we don't know if it's done.
            unload = False

        # Let context decide when to run this.
        if context is not None:
            # Processing is done, let context know.
            return context.run_script_generic(script_name)

        logger.debug("Running script: %s", js)

        script = self.session.create_script(js, runtime=runtime)

        script.on('message', on_message)

        try:
            script.load()
        except:
            logger.error("Error running script!")
            script.unload()
            return

        # If we're async, wait until we're done
        # This needs to happen before unlink so we ensure we get all messages
        while onComplete is not None and completed == []:
            time.sleep(0.01)
        
        if unload:
            script.unload()
        else:
            # Inserting instead of appending since earlier scripts need to be unloaded later
            self._scripts.insert(0, [script, js])


        return msg, data

    def _resolve_location_string(self, location):
        """Take location string and resolve it into an integer address."""
        #assert type(location) is str, "Invalid call to resolve_location_string with type {}".format(type(location))
        if isinstance(location, int):
            return types.Pointer(location)

        return self.modules.lookup_symbol(location)

    def __repr__(self):
        attrs = ['Process', self.file_name + ":" + str(self.pid)]
        return '<' + ' '.join(attrs) + '>'

    ############
    # Property #
    ############

    @property
    def device_platform(self):
        """Wrapper to discover the device's platform."""

        def message(x, y):
            self.device_platform = x['payload']

        try:
            return self.__device_platform
        except:
            pass

        js = "send(Process.platform)"
        script = self.session.create_script(js)
        script.on('message', message)
        script.load()
        return self.__device_platform

    @device_platform.setter
    def device_platform(self, platform):
        self.__device_platform = platform

    @property
    def pid(self):
        return self.session._impl.pid

    @property
    def entrypoint(self):
        """int: Returns the entrypoint for this running program."""
        mod = self.modules[self.file_name]

        if self.__entrypoint is None:
            if self.file_type == 'ELF':
                self.__entrypoint = types.Pointer(int(self.run_script_generic("""send(Memory.readPointer(ptr(Number(Process.getModuleByName('{}').base) + 0x18)))""".format(self.file_name), raw=True, unload=True)[0][0],16))
                
                if mod.elf.type_str == 'DYN':
                    self.__entrypoint = types.Pointer(self.__entrypoint + mod.base)

            else:
                logger.warn('entrypoint not implemented for file of type {}'.format(self.file_type))
                return None
            
        # TODO: Windows?
        # TODO: Mac?

        return self.__entrypoint

    @property
    def endianness(self):
        """Determine which endianness this binary is. (little, big)"""

        if self.__endianness != None:
            return self.__endianness

        if self.device_platform == 'windows':
            # TODO: Technically assumption, but like 99% of the time it's right.
            self.__endianness = 'little'

        elif self.file_type == 'ELF':
            endianness = self.run_script_generic("""send(ptr(Number(Process.enumerateModulesSync()[0].base) + 5).readS8())""", raw=True, unload=True)[0][0]
            self.__endianness = 'little' if endianness == 1 else 'big'

        else:
            logger.warn("Unhandled endianness check for ({}, {}), assuming little".format(self.file_type, self.device_platform))

        return self.__endianness

    @property
    def file_type(self):
        """Guesses the file type."""

        # TODO: Android processes we attach to can't getModuleByName for their file name...
        # Maybe use "app_process64" instead... Or resolve the first module and go with that..
        if isinstance(self.device, device_types.AndroidDevice):
            return "ELF"

        # TODO: Update this with other formats. PE/COFF/MACHO/etc
        if self.__file_type is None:
            if self.run_script_generic("""send('bytes', Process.getModuleByName('{}').base.readByteArray(4))""".format(self.file_name), raw=True, unload=True)[1][0] == b'\x7fELF':
                self.__file_type = 'ELF'
            elif self.run_script_generic("""send('bytes', Process.getModuleByName('{}').base.readByteArray(2))""".format(self.file_name), raw=True, unload=True)[1][0] == b'MZ':
                self.__file_type = "PE"
            else:
                self.__file_type = 'Unknown'

        return self.__file_type

    @property
    def file_name(self):
        """str: The base file name."""
        # TODO: This assumes the base module is always first...
        if self.__file_name is None:
            self.__file_name = self.run_script_generic("""send(Process.enumerateModulesSync())""", raw=True, unload=True)[0][0][0]['name']

        return self.__file_name

    @property
    def bits(self):
        """int: How many bits is the CPU?"""
        if self.__bits == None:
            self.__bits = self.run_script_generic("""send(Process.pointerSize);""", raw=True, unload=True)[0][0] * 8
        
        return self.__bits

    @property
    def arch(self):
        """str: What architecture? (x64, ia32, arm, others?)"""
        try:
            return self.__arch
        except:
            known_arch = ['x64', 'ia32', 'arm']
            arch = self.run_script_generic("""send(Process.arch);""", raw=True, unload=True)[0][0]

            if arch not in known_arch:
                raise Exception("Unknown arch returned from Frida: {}".format(arch))
            self.__arch = arch
            return self.__arch

    @property
    def verbose(self):
        """bool: Output extra debugging information."""
        return self.__verbose

    @verbose.setter
    def verbose(self, verbose):
        if verbose:
            logging.getLogger().setLevel(logging.DEBUG)
            #logger.setLevel(logging.DEBUG)
        self.__verbose = verbose

    @property
    def target(self):
        """str, int: Target for this session."""
        return self.__target

    @target.setter
    def target(self, target):
        # Check if this is a pid
        try:
            p = next(x for x in self.device.device.enumerate_processes() if x.pid == common.auto_int(target))
            target = p.pid
            self.__file_name = p.name

        except (StopIteration, ValueError):
            pass

        if isinstance(target, str):
            full_path = os.path.abspath(target)
            self.__file_name = os.path.basename(full_path)

            if os.path.isfile(full_path):
                self._spawn_target = full_path
            

        self.__target = target

    @property
    def alive(self):
        """bool: Is this process still alive?"""
        try:
            next(True for x in self.device.device.enumerate_processes() if x.pid == self.pid)
            return True
        except StopIteration:
            return False

    @property
    def device(self):
        """frida.core.Device: Frida device object for this connection."""
        return self.__device

    @device.setter
    def device(self, device):
        assert isinstance(device, device_types.BaseDevice), "Device must be an instantiation of one of the devices defined in revenge.device_types."
        self.__device = device
        """
        if isinstance(device, device_types.LocalDevice):
            self.__device = device.device

        elif isinstance(device, device_types.AndroidDevice):
            self.__device = device.device

        else:
            error = "Unexpected/unhandled device type of {}".format(type(device))
            logger.error(error)
            raise Exception(error)
        """
    
    @property
    def BatchContext(self):
        """Returns a BatchContext class for this process.

        Example:
            .. code-block:: python3

                with process.BatchContext() as context:
                    something(context=context)

        """
        return lambda *args, **kwargs: BatchContext(self, *args, **kwargs)


from . import common, types, config, device_types
from .memory import Memory
from .threads import Threads
from .tracer import Tracer
from .modules import Modules
from .java import Java
from .contexts import BatchContext

# Doc fixups
Process.BatchContext.__doc__ += BatchContext.__init__.__doc__


def sigint_handler(sig, frame):
    sys.exit()

signal.signal(signal.SIGINT, sigint_handler)

def main():
    signal.signal(signal.SIGINT, sigint_handler)

    global process
    process = Process()

    while True:
        time.sleep(1)

if __name__ == '__main__':
    main()




import logging
logger = logging.getLogger(__name__)

from prettytable import PrettyTable
import binascii
import operator
import struct
from termcolor import cprint, colored

from .. import common, types, symbols

class Memory(object):
    """Class to simplify getting and writing things to memory. Behaves like a list.
    
    Example:
        - memory[0x12345].int8 -> Reads a signed 8-bit int from address
        - memory[0x12345:0x12666] -> Returns byte array from memory
    """

    def __init__(self, process):
        self._process = process

        # Keep track of where we've inserted breakpoints
        # key == address of breakpoint, value == memory location to un-breakpoint it
        self._active_breakpoints = {}

        # Keep track of what we've allocated.
        # key == address of allocation, value = script where we allocated it.
        # NOTE: It's important to keep the script alive until we're done with the alloc or javascript might gc it.
        self._allocated_memory = {}

        # key == address of replaced function, value = tuple: what it's being replaced with, script so we can unload later
        self._active_replacements = {}


    def alloc(self, size):
        """Allocate size bytes of memory and get a MemoryBytes object back to use it.
    
        Args:
            size (int): How many bytes to allocate.
        """
        
        assert type(size) is int

        pointer = common.auto_int(self._process.run_script_generic("""var p = Memory.alloc(uint64('{}')); send(p);""".format(hex(size)), raw=True, unload=False)[0][0])
        script = self._process._scripts.pop(0) # We want to hold on to it here

        self._allocated_memory[pointer] = script
        return MemoryBytes(self._process, pointer, pointer+size)

    def alloc_string(self, s, encoding='latin-1'):
        """Short-hand to run alloc of appropriate size, then write in the string.
        
        Args:
            s (bytes, str): String to allocate
            encoding (str, optional): How to encode the string if passed in as type str.
        """

        # TODO: Smart guess encoding, linux is usually utf-8, Windows has function call to determine utf-8 vs 16. Mac...?

        if type(s) in [types.StringUTF8, types.StringUTF16]:
            if s.type == 'utf8':
                encoding = 'utf-8'
            elif s.type == 'utf16':
                encoding = 'utf-16'
            else:
                logger.error('How did i get here??')
                return

            s = str(s)
        
        if type(s) is str:
            s = s.encode(encoding)
            if encoding == 'utf-16':
                s = s[2:] # Remove BOM
                s += b'\x00' # Extra null at end of utf-16
        

        if type(s) is not bytes:
            logger.error("Invalid string type of {}".format(type(s)))
            return None
        
        # Null terminate
        s += b'\x00'

        mem = self.alloc(len(s))
        mem.bytes = s
        return mem

    def find(self, *args, **kwargs):
        """Search for thing in memory. Must be one of the defined types."""
        return MemoryFind(self._process, *args, **kwargs)

    def describe_address(self, address, color=False):
        """Takes in address and attempts to return a better description of what's there.
        
        Args:
            address (int): What address to describe
            color (bool, optional): Should the description be colored?
                (default: False)

        Returns:
            str: description of the address
        """

        if isinstance(address, symbols.Symbol):
            address = address.address

        assert isinstance(address, int), "Unexpected address type of {}".format(type(address))

        desc = ""
        module = self._process.modules[address]

        if module is not None:
            if color:
                desc += colored(module.name,"magenta") or ""
            else:
                desc += module.name or ""

            try:
                # If we can find a closest function, use that.
                func_name, func_addr = next((name, addr.address) for name,addr in sorted(module.symbols.items(), key=operator.itemgetter(1),reverse=True) if address >= addr)
                func_name = str(func_name)

                offset = address - func_addr

                # This is probably not really the function
                if func_name.startswith('plt.') and offset >= 0x10:
                    raise StopIteration

                if color:
                    desc += ":" + colored(func_name, "magenta", attrs=["bold"])
                else:
                    desc += ":" + func_name

            except StopIteration:
                # We did not find a closest function, just offset from module base
                offset = address - module.base

            if offset != 0:
                if color:
                    desc += "+" + colored(hex(offset), "cyan")
                else:
                    desc += "+" + hex(offset)

        else:
            if color:
                desc += colored(hex(address), "cyan")
            else:
                desc += hex(address)

        return desc

    def _type_to_search_string(self, thing):
        """Converts the given object into something relevant that can be fed into a memory search query."""

        if not isinstance(thing, types.all_types):
            logger.error("Please use valid type.")
            return None

        endian_str = "<" if self._process.endianness == 'little' else '>'

        if isinstance(thing, types.StringUTF8):
            # Normal string
            return binascii.hexlify(thing.encode('utf-8')).decode()

        elif isinstance(thing, types.StringUTF16):
            # Wide Char String (Windows/UTF16)
            return binascii.hexlify(thing.encode('utf-16')[2:]).decode()

        elif isinstance(thing, types.UInt8):
            return binascii.hexlify(struct.pack(endian_str + "B", thing)).decode()

        elif isinstance(thing, types.Int8):
            return binascii.hexlify(struct.pack(endian_str + "b", thing)).decode()

        elif isinstance(thing, types.UInt16):
            return binascii.hexlify(struct.pack(endian_str + "H", thing)).decode()

        elif isinstance(thing, types.Int16):
            return binascii.hexlify(struct.pack(endian_str + "h", thing)).decode()

        elif isinstance(thing, types.UInt32):
            return binascii.hexlify(struct.pack(endian_str + "I", thing)).decode()

        elif isinstance(thing, types.Int32):
            return binascii.hexlify(struct.pack(endian_str + "i", thing)).decode()

        elif isinstance(thing, types.UInt64):
            return binascii.hexlify(struct.pack(endian_str + "Q", thing)).decode()

        elif isinstance(thing, types.Int64):
            return binascii.hexlify(struct.pack(endian_str + "q", thing)).decode()
        
        else:
            logger.error("Unexpected type to convert of {}".format(type(thing)))
            return None
        

    def __getitem__(self, item):

        if type(item) == str:
            # Assume it's something we need to resolve
            item = self._process._resolve_location_string(item)

        if isinstance(item, symbols.Symbol):
            item = item.address

        if isinstance(item, int):
            return MemoryBytes(self._process, item)

        elif type(item) == slice:

            if item.start is None or item.stop is None or item.step is not None:
                logger.error("Memory slices must have start and stop and not contain a step option.")
                return

            return MemoryBytes(self._process, item.start, item.stop)

        logger.error("Unhandled memory type of {}".format(type(item)))

    @property
    def maps(self):
        """Return a list of memory ranges that are currently allocated."""
        return MemoryMap(self._process)

    def __str__(self):
        
        table = PrettyTable(['range', 'prot', 'file'])
        table.header = False
        table.align = 'l'
        table.border = False

        for range in self.maps:
            table.add_row([
                hex(range.base)[2:] + '-' + hex(range.base+range.size)[2:],
                range.protection,
                range.file or '',
                ])

        return str(table)

from . import MemoryBytes
from . import MemoryRange
from . import MemoryFind
from . import MemoryMap

Memory.find.__doc__ = MemoryFind.__init__.__doc__




import logging
logger = logging.getLogger(__name__)

from fnmatch import fnmatch

class AndroidApplications:
    def __init__(self, android):
        self._android = android

    def __iter__(self):
        return self._android.device.enumerate_applications().__iter__()

    def __len__(self):
        return len(list(self.__iter__()))

    def __getitem__(self, item):
        if isinstance(item, int):
            return list(self)[item]

        elif isinstance(item, str):
            match = [x for x in self if fnmatch(x.identifier, item) or fnmatch(x.name, item)]
            if len(match) > 1:
                return match
            if match == []:
                return None
            return match[0]

        else:
            logger.error("Unhandled item get of type {}".format(type(item)))
            return

    def __repr__(self):
        attrs = ["AndroidApplications", str(len(self)), "installed"]

        return "<" + " ".join(attrs) + ">"




from .. import Colorer
import logging
logging.basicConfig(level=logging.WARN)
logger = logging.getLogger(__name__)

import sys

import colorama
colorama.init()
from termcolor import cprint, colored

import argparse
import revenge

common = revenge.common
types = revenge.types
Process = revenge.Process

from . import actions

def parse_args():

    parser = argparse.ArgumentParser(
        description='CLI wrapper around Frida-Util.'
        )

    parser.add_argument('--tid', type=int, default=None,
            help="Thread to stalk. (Default: all threads.)")
    parser.add_argument('--include-module', "-I", type=str, default=None, metavar='module', nargs='+',
            help="Module to include for stalking (default: All modules).")
    parser.add_argument('--include-function', "-i", type=str, default=None, metavar='module:offset',
            help="Function to include for stalking (default: All functions).")
    parser.add_argument('--replace-function', "-rf", type=str, default=[], metavar='<module:offset|symbol>?<return_val>', nargs='+',
            help="Replace given function by simply returning the given value instead.")
    parser.add_argument('--pause-at', type=str, default=[], metavar='<module:offset|symbol>', nargs='+',
            help="Pause execution at address.")
    parser.add_argument('--verbose', "-v", action='store_true', default=False,
            help="Output more verbose information (defualt: False)")
    parser.add_argument('--rw-everything', '-rw', default=False, action='store_true',
            help="Change all r-- memory areas into rw-. This can sometimes help segfault issues (default: off)")

    stalk_group = parser.add_argument_group('stalk options')
    stalk_group.add_argument('--call', action='store_true', default=False,
            help="Stalk calls")
    stalk_group.add_argument('--ret', action='store_true', default=False,
            help="Stalk rets")
    stalk_group.add_argument('--exec', action='store_true', default=False,
            help="Stalks every single instruction.")
    stalk_group.add_argument('--block', action='store_true', default=False,
            help="Stalks every code block.")
    stalk_group.add_argument('--compile', action='store_true', default=False,
            help="Stalks every time Frida needs to compile.")

    windows_group = parser.add_argument_group('windows options')
    windows_group.add_argument('--windows-message', '-wm', default=None, type=str, nargs='+', metavar='Message',
            help="Down select to these specific windows messages (i.e.: WM_KEYUP, WM_KEYDOWN).")

    find_group = parser.add_argument_group('find options')
    find_group.add_argument('--string', type=str, default=None,
            help="Search for string in program memory.")
    find_group.add_argument('--uint8', type=int, default=None,
            help="Search for unsigned 8bit int in program memory.")
    find_group.add_argument('--int8', type=int, default=None,
            help="Search for signed 8bit int in program memory.")
    find_group.add_argument('--uint16', type=int, default=None,
            help="Search for unsigned 16bit int in program memory.")
    find_group.add_argument('--int16', type=int, default=None,
            help="Search for signed 16bit int in program memory.")
    find_group.add_argument('--uint32', type=int, default=None,
            help="Search for unsigned 32bit int in program memory.")
    find_group.add_argument('--int32', type=int, default=None,
            help="Search for signed 32bit int in program memory.")
    find_group.add_argument('--uint64', type=int, default=None,
            help="Search for unsigned 64bit int in program memory.")
    find_group.add_argument('--int64', type=int, default=None,
            help="Search for signed 64bit int in program memory.")
    find_group.add_argument('--number', type=int, default=None,
            help="Search for number of any size in program memory.")

    spawn_group = parser.add_argument_group('spawn options')
    spawn_group.add_argument('--file', '-f', type=str, metavar=('FILE','ARGS'), default=None, nargs='+',
            help="Spawn file.")
    spawn_group.add_argument('--resume', default=False, action='store_true',
            help="Resume binary after spawning it (default: false).")

    parser.add_argument('action', choices=('stalk', 'windows_messages', 'find', 'diff_find', 'ipython'),
            help="What type of stalking.")

    parser.add_argument('target', type=str,
            help="Target to attach to.")

    args = parser.parse_args()

    # Clean up windows messages
    if args.windows_message is not None:
        args.windows_message = [common.windows_messages_by_name[x] for x in self._args.windows_message]

    return args

def replace_function(process, f):
    """Replace a given function to always return a given value. <module:offset|symbol>?<return_val>"""
    assert type(f) == str, "Unexpected replace function argument type of {}".format(type(f))

    location, return_value = f.split("?")
    replace_location = process._resolve_location_string(location)

    replace_vars = {
            "FUNCTION_RETURN_VALUE_HERE": return_value,
            "FUNCTION_ADDRESS_HERE": hex(replace_location),
            }

    self.run_script_generic("replace_function.js", replace=replace_vars)

def main():
    args = parse_args()

    process = Process(args.target, resume=args.resume, verbose=args.verbose)

    if args.rw_everything:
        print("RW'ing memory areas\t\t... ", end='', flush=True)
        process.run_script_generic('rw_everything.js', unload=True)
        cprint('[ DONE ]', 'green')


    for f in args.replace_function:
        location, return_value = f.split("?")
        mem = process.memory[location]
        mem.replace = int(return_value, 0)

    # Setup any requested pauses
    for location in args.pause_at:
        process.memory[location].breakpoint = True

    if args.action == 'ipython':
        import IPython
        IPython.embed()

    elif args.action == 'find':
        action_find = actions.ActionFind(process, **vars(args))
        action_find.run()
        print({hex(x):y for x,y in action_find.discovered_locations.items()})
        
    elif args.action == 'stalk':
        action_stalker = actions.ActionStalker(process, **vars(args))
        action_stalker.run()

    """
    elif self._args.action == 'windows_messages':
        self.action_windows_messages = actions.ActionWindowsMessages(self, **vars(self._args))
        self.action_windows_messages.run()


    elif self._args.action == 'diff_find':
        self.action_diff = actions.ActionDiffFind(self, **vars(self._args))
        self.action_diff.run()
    
    time.sleep(1)
    """

    """
    # Make sure the requested include module exists
    # TODO: Move this into UI module once i make it..
    if self._args.include_module is not None:
        try:
            bad_mod = next(module for module in self._args.include_module if module not in self.modules)
            logger.warn("Your chosen include_module ({}) doesn't match any modules found. Double check the capitalization or spelling.".format(bad_mod))
        except:
            pass
    """

if __name__ == '__main__':
    main()




import logging
logger = logging.getLogger(__name__)

import json

from .. import common, types
import time

class MemoryFind(object):

    def __init__(self, process, thing, ranges=None):
        """Find something in memory.

        Args:
            process: Base process instantiation
            thing: Some instantiated type to search for from types module
            ranges(list, optional): List of MemoryRange objects to limit the search
                to. By default, search everything.
        """
        self._process = process
        self.thing = thing
        self.ranges = ranges

        # Loaded script to unload once we're done
        self._script = None

        # Not completed yet
        self.completed = False

        # Memory locations we discovered
        self.found = set()

        self._start()

    def sleep_until_completed(self):
        """This call sleeps and only returns once the search is completed."""

        while not self.completed:
            time.sleep(0.1)

    def _start(self):
        """Starts the search."""

        replace = {
                "SCAN_PATTERN_HERE": self.search_string,
                "SEARCH_SPACE_HERE": json.dumps(self._ranges_js),
                }

        self._process.run_script_generic("find_in_memory.js", replace=replace, unload=False, on_message=self._on_message)
        self._script = self._process._scripts.pop(0)

    def _on_message(self, m,d):
        """Catch messages from our search."""
        payload = m['payload']

        if type(payload) is list:
            for addr in payload:
                self.found.add(types.Pointer(common.auto_int(addr['address'])))
        
        elif type(payload) is str and payload == 'DONE':
            self.completed = True

        else:
            logger.error("Unexpected message: {} {}".format(m,d))

    def __del__(self):
        # Be sure to unload our script
        if self._script is not None:
            self._script[0].unload()
            self._script = None

    def __repr__(self):
        attr = ["MemoryFind"]
        attr += ["found", str(len(self.found))]
        
        if self.completed:
            attr.append("completed")
        else:
            attr.append("running")

        return "<{}>".format(' '.join(attr))

    def __iter__(self):
        if not self.completed:
            logger.warning("Search is not completed. Your results may not be full.")

        return (x for x in self.found)

    def __len__(self):
        return len(self.found)

    @property
    def completed(self):
        """bool: Is this search completed?"""
        return self.__completed

    @completed.setter
    def completed(self, completed):
        assert type(completed) is bool
        self.__completed = completed

        # Clean up our search script when we're done.
        if completed and self._script is not None:
            self._script[0].unload()
            self._script = None

    @property
    def search_string(self):
        """The search string for this thing."""
        return self._process.memory._type_to_search_string(self.thing)

    @property
    def thing(self):
        """What we're looking for."""
        return self.__thing

    @thing.setter
    def thing(self, thing):
        
        if not isinstance(thing, types.all_types):
            error = "Invalid search thing of type {}".format(type(thing))
            logger.error(error)
            raise Exception(error)

        self.__thing = thing

    @property
    def _ranges_js(self):
        """Returns the ranges as a list for insertion into js."""
        l = []
        for range in self.ranges:
            d = {'base': range.base.js, 'size': range.size}
            l.append(d)

        return l

    @property
    def ranges(self):
        return self.__ranges

    @ranges.setter
    def ranges(self, ranges):
        
        if ranges is None:
            # It appears the first time maps gets run, something in the Frida actually changes... Not sure what.
            # Running this here to prime the pump as it were.. Maybe some day figure out wtf is going on.
            # REMINDER: This bug didn't always hit. So pytest may say it's fine when it isn't.
            self._process.memory.maps
            self._process.memory.maps

            ranges = list(self._process.memory.maps)

        if type(ranges) is MemoryRange:
            self.__ranges = [ranges]

        if type(ranges) not in (list, tuple):
            error = "Invalid range type of {}".format(type(ranges))
            logger.error(error)
            raise Exception(error)

        self.__ranges = ranges

from . import MemoryRange



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
import os, sys, ast

here = os.path.abspath(os.path.dirname(__file__))
version = '0.11'

#with open(path.join(here, 'README.md'), encoding='utf-8') as f:
#    long_description = f.read()
long_description = "See website for more info."

# Frida 12.6.11 -> https://github.com/frida/frida/issues/986

setup(
    name='revenge',
    version=version,
    description='REVerse ENGineering Environment',
    long_description=long_description,
    url='https://github.com/bannsec/revenge',
    author='Michael Bann',
    author_email='self@bannsecurity.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Operating System :: POSIX :: Linux',
        'Environment :: Console'
    ],
    keywords='frida python3 reversing dbi',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=['frida', 'prettytable', 'colorama', 'termcolor', 'psutil', 'pyelftools', 'pefile', 'appdirs', 'bs4', 'requests'],
    extras_require={
        'dev': ['ipython','twine','pytest','python-coveralls','coverage','pytest-cov','pytest-xdist','sphinxcontrib-napoleon', 'sphinx_rtd_theme','sphinx-autodoc-typehints', 'pyOpenSSL', 'numpy'],
    },
    entry_points={
        'console_scripts': [
            'revenge = revenge.cli.cli:main',
        ],
    },
    include_package_data = True,
)





import logging
logger = logging.getLogger(__name__)

import json
from .. import common, types

class MemoryBytes(object):
    """Meta-class used for resolving bytes into something else."""

    def __init__(self, process, address, address_stop=None):
        """Abstracting what memory location is.

        Args:
            process: Util object
            address (int): Starting address of the memory location.
            address_stop (int, optional): Optional stopping memory location.
        """
        self._process = process
        self.address = address
        self.address_stop = address_stop
        self.return_type = types.Pointer # Default

    def free(self):
        """bool: Free this memory location. This is only valid if this memory location has been allocated by us."""

        # Make sure we allocated it
        if self.address not in self._process.memory._allocated_memory:
            logger.error("Can't free this memory as we didn't allocate it.")
            return False

        # Free it implicitly by freeing our script
        script = self._process.memory._allocated_memory.pop(self.address)
        script[0].unload()
        return True

    def __repr__(self):
        attrs = ['MemoryBytes', hex(self.address)]

        if self.size is not None:
            attrs.append(str(self.size) + ' bytes')

        if self.replace is not None:
            attrs.append("Replaced")

        return "<{}>".format(' '.join(attrs))

    def __call__(self, *args, **kwargs):
        """Call this memory location as a function.
        
        *args will be parsed and passed to the actual function
        **kwargs will be passed to Process.run_script_generic
        """

        # Generically use pointers and figure it out later

        # Resolve args to memory strings and such if needed
        args_resolved = []
        to_free = []
        args_types = []

        for arg in args:

            if type(arg) is MemoryBytes:
                arg = arg.address

            # Make temporary string first
            if type(arg) in [types.StringUTF16, types.StringUTF8]:
                s = self._process.memory.alloc_string(arg)
                args_resolved.append('ptr("' + hex(s.address) + '")')
                to_free.append(s)
                args_types.append('pointer')

            # Make temporary string in memory
            elif type(arg) in [str, bytes]:
                s = self._process.memory.alloc_string(arg)
                args_resolved.append('ptr("' + hex(s.address) + '")')
                to_free.append(s)
                args_types.append('pointer')

            elif type(arg) is int:
                # Defaulting these to pointers for now.
                args_resolved.append('ptr("' + hex(arg) + '")')
                args_types.append('pointer')

            elif isinstance(arg, types.all_types):
                args_resolved.append(arg.js)
                args_types.append(arg.type)

            else:
                logger.error("Unexpected argument type of {}".format(type(arg)))
                return None

        js = """var f = new NativeFunction(ptr("{ptr}"), '{ret_type}', {args_types});""".format(
                ptr = hex(self.address),
                ret_type = self.return_type.type,
                args_types = json.dumps(args_types),
            )
        
        # If we are not passing to a context, then sync send it
        if kwargs.get("context", None) is None:
            js += "send(f({args}))".format(args = ', '.join(args_resolved))
        else:
            # We're passing to a context, let it handle the message back.
            js += "f({args})".format(args = ', '.join(args_resolved))

        # Wrap call to watch for native exceptions
        js = "try { " + js + """} catch (exception) { 
            if ( Object.keys(exception).indexOf('memory') == -1 ) {
                var bt = [];
            } else {
                var bt = Thread.backtrace(exception['memory']['context']);
            }
            send({
                "exception": exception,
                "backtrace": bt,
            })}"""

        ret = self._process.run_script_generic(js, raw=True, unload=True, **kwargs)

        # If we changed on_message or context, this might be None. That's ok.
        if ret is None:
            ret = 0
        else:
            ret = ret[0][0]

        # Free stuff up
        for alloc in to_free:
            # If we dynamically allocated something but we're in a context, we
            # cannot free it yet.  Warn the user.
            if kwargs.get("context", None) is not None:
                logger.warning("Not freeing dynamically allocated memory due to use of context. This will cause a memory leak!!")
            else:
                alloc.free()

        # Handle the case where the process did something bad
        if isinstance(ret, dict) and "exception" in ret:
            return NativeException._from_frida_dict(self._process, ret['exception'], ret['backtrace'])
        
        return self.return_type(common.auto_int(ret))

    def _remove_replace(self):
        """Reverts any replacement of this function."""

        if self.address in self._process.memory._active_replacements:
            self._process.memory._active_replacements[self.address][1][0].unload()
            self._process.memory._active_replacements.pop(self.address)

    @property
    def replace_on_message(self):
        """callable: Optional callable to be called if/when something inside the function replace sends data back.
        
        Example:
            .. code-block:: python3

                # If you just wanted to print out the messages that came back
                def on_message(x,y):
                    print(x,y)

                strlen.replace_on_message = on_message
        """

        try:
            return self.__replace_on_message
        except AttributeError:
            return None

    @replace_on_message.setter
    def replace_on_message(self, replace_on_message):

        if replace_on_message is None:
            self.__replace_on_message = None
            return

        if not callable(replace_on_message):
            logger.error("On Message handler must be callable.")
            return

        self.__replace_on_message = replace_on_message

        # force reload the modification with the new on_message handler
        self.replace = self.replace

    @property
    def replace(self):
        """What is this function being replaced by? None if there's no replacement.
        
        Examples:
            
            Replacing strlen with a function that sends back the argument

            .. code-block:: python3

                strlen = process.memory[':strlen']
                strlen.return_type = types.Int64
                strlen.argument_types = types.Pointer

                # If you're not sending messages, you don't need this
                def on_message(x,y):
                    print(x,y)

                strlen.replace_on_message = on_message

                # Replacing strlen with some arbitrary js function
                # "original" is always going to be the function you're replacing
                # In this case, take response and decrement it by one
                strlen.replace = "function (s) { send(s); return original(s)-1;}"
                assert strlen("hello") == 4

                # Remove the replace
                strlen.replace = None
                assert strlen("test") == 4

            Replacing alarm to return 1

            .. code-block:: python3

                alarm = process.memory[':alarm']
                # Replace function by just returning static value
                alarm.replace = 1

                # Alarm is not set
                assert alarm(1) == 1
        """
        try:
            return self._process.memory._active_replacements[self.address][0]
        except:
            return None

    @replace.setter
    def replace(self, replace):

        self._remove_replace()

        if replace is None:
            return

        #
        # Replace function with simple return value
        #

        if isinstance(replace, int):

            # If it's not already a defined type
            if type(replace) is int:
                replace = self.return_type(replace)
                replace_val = replace.js
                replace_type = self.return_type.type

            else:
                replace_val = replace.js
                replace_type = replace.type

            replace_func = """function () {{ return {}; }}""".format(replace_val)

            replace_vars = {
                "FUNCTION_RETURN_TYPE": replace_type,
                "FUNCTION_ADDRESS": self.address.js,
                "FUNCTION_REPLACE": replace_func,
                "FUNCTION_ARG_TYPES": str([])
            }

            self._process.run_script_generic("replace_function.js", replace=replace_vars, unload=False)
            script = self._process._scripts.pop(0)
            self._process.memory._active_replacements[self.address] = (replace, script)

        #
        # Replace function with js
        #

        elif isinstance(replace, str):

            # Default return to pointer
            return_type = self.return_type.type or types.Pointer

            # Default to no argument types
            arg_types = [x.type for x in self.argument_types] if self.argument_types is not None else []
            arg_types = str(arg_types)

            replace_vars = {
                "FUNCTION_RETURN_TYPE": return_type,
                "FUNCTION_ADDRESS": self.address.js,
                "FUNCTION_REPLACE": replace,
                "FUNCTION_ARG_TYPES": arg_types
            }

            self._process.run_script_generic("replace_function.js",
                    replace=replace_vars, unload=False,
                    on_message=self.replace_on_message)
            script = self._process._scripts.pop(0)
            self._process.memory._active_replacements[self.address] = (replace, script)

        else:
            logger.error("Invalid replacement type of {}".format(type(replace)))

    @property
    def implementation(self):
        return self.replace

    @implementation.setter
    def implementation(self, implementation):
        self.replace = implementation

    @property
    def argument_types(self):
        """tuple: Returns the registered arguments types for this function or
        None if none have been found/registered."""

        try:
            return self.__argument_types
        except AttributeError:
            return None

    @argument_types.setter
    def argument_types(self, arg_types):
        
        if arg_types is None:
            self.__argument_types = None
            return

        if isinstance(arg_types, list):
            arg_types = tuple(arg_types)

        if not isinstance(arg_types, tuple):
            arg_types = (arg_types,)

        if not all(t in types.all_types for t in arg_types):
            logger.error("All argument types must be valid revenge.type types.")
            return

        self.__argument_types = arg_types

        # force reload the modification with the new on_message handler
        self.replace = self.replace

    @property
    def return_type(self):
        """What's the return type for this? Only valid if this is a function."""
        return self.__return_type

    @return_type.setter
    def return_type(self, ret):

        if type(ret) is not type:
            logger.error('Please set with types.<type>.')
            return

        if ret not in types.all_types:
            logger.error('Unexpected type of {}. Please use types.<type>.'.format(ret))
            return

        self.__return_type = ret

        # force reload the modification with the new on_message handler
        self.replace = self.replace

    @property
    def int8(self):
        """Signed 8-bit int"""
        return types.Int8(self._process.run_script_generic("""send(ptr("{}").readS8())""".format(hex(self.address)), raw=True, unload=True)[0][0])

    @int8.setter
    def int8(self, val):
        self._process.run_script_generic("""send(ptr("{}").writeS8({}))""".format(hex(self.address), val), raw=True, unload=True)[0][0]

    @property
    def uint8(self):
        """Unsigned 8-bit int"""
        return types.UInt8(self._process.run_script_generic("""send(ptr("{}").readU8())""".format(hex(self.address)), raw=True, unload=True)[0][0])

    @uint8.setter
    def uint8(self, val):
        self._process.run_script_generic("""send(ptr("{}").writeU8({}))""".format(hex(self.address), val), raw=True, unload=True)[0][0]

    @property
    def int16(self):
        """Signed 16-bit int"""
        return types.Int16(self._process.run_script_generic("""send(ptr("{}").readS16())""".format(hex(self.address)), raw=True, unload=True)[0][0])

    @int16.setter
    def int16(self, val):
        self._process.run_script_generic("""send(ptr("{}").writeS16({}))""".format(hex(self.address), val), raw=True, unload=True)[0][0]

    @property
    def uint16(self):
        """Unsigned 16-bit int"""
        return types.UInt16(self._process.run_script_generic("""send(ptr("{}").readU16())""".format(hex(self.address)), raw=True, unload=True)[0][0])

    @uint16.setter
    def uint16(self, val):
        self._process.run_script_generic("""send(ptr("{}").writeU16({}))""".format(hex(self.address), val), raw=True, unload=True)[0][0]

    @property
    def int32(self):
        """Signed 32-bit int"""
        return types.Int32(self._process.run_script_generic("""send(ptr("{}").readS32())""".format(hex(self.address)), raw=True, unload=True)[0][0])

    @int32.setter
    def int32(self, val):
        self._process.run_script_generic("""send(ptr("{}").writeS32({}))""".format(hex(self.address), val), raw=True, unload=True)[0][0]

    @property
    def uint32(self):
        """Unsigned 32-bit int"""
        return types.UInt32(self._process.run_script_generic("""send(ptr("{}").readU32())""".format(hex(self.address)), raw=True, unload=True)[0][0])

    @uint32.setter
    def uint32(self, val):
        self._process.run_script_generic("""send(ptr("{}").writeU32({}))""".format(hex(self.address), val), raw=True, unload=True)[0][0]

    @property
    def int64(self):
        """Signed 64-bit int"""
        return types.Int64(common.auto_int(self._process.run_script_generic("""send(ptr("{}").readS64())""".format(hex(self.address)), raw=True, unload=True)[0][0]))

    @int64.setter
    def int64(self, val):
        self._process.run_script_generic("""ptr("{}").writeS64(int64('{}'))""".format(hex(self.address), val), raw=True, unload=True)
    
    @property
    def uint64(self):
        """Unsigned 64-bit int"""
        return types.UInt64(common.auto_int(self._process.run_script_generic("""send(ptr("{}").readU64())""".format(hex(self.address)), raw=True, unload=True)[0][0]))

    @uint64.setter
    def uint64(self, val):
        self._process.run_script_generic("""ptr("{}").writeU64(uint64('{}'))""".format(hex(self.address), val), raw=True, unload=True)

    @property
    def string_ansi(self):
        """Read as ANSI string"""
        return self._process.run_script_generic("""send(ptr("{}").readAnsiString())""".format(hex(self.address)), raw=True, unload=True)[0][0]

    @string_ansi.setter
    def string_ansi(self, val):
        self._process.run_script_generic("""send(ptr("{}").writeAnsiString(\"{}\"))""".format(hex(self.address), val), raw=True, unload=True)

    @property
    def string_utf8(self):
        """Read as utf-8 string"""
        return self._process.run_script_generic("""send(ptr("{}").readUtf8String())""".format(hex(self.address)), raw=True, unload=True)[0][0]

    @string_utf8.setter
    def string_utf8(self, val):
        self._process.run_script_generic("""send(ptr("{}").writeUtf8String(\"{}\"))""".format(hex(self.address), val), raw=True, unload=True)

    @property
    def string_utf16(self):
        """Read as utf-16 string"""
        return self._process.run_script_generic("""send(ptr("{}").readUtf16String())""".format(hex(self.address)), raw=True, unload=True)[0][0]

    @string_utf16.setter
    def string_utf16(self, val):
        self._process.run_script_generic("""send(ptr("{}").writeUtf16String(\"{}\"))""".format(hex(self.address), val), raw=True, unload=True)

    @property
    def double(self):
        """Read as double val"""
        return types.Double(self._process.run_script_generic("""send(ptr("{}").readDouble())""".format(hex(self.address)), raw=True, unload=True)[0][0])

    @double.setter
    def double(self, val):
        self._process.run_script_generic("""send(ptr("{}").writeDouble({}))""".format(hex(self.address), val), raw=True, unload=True)

    @property
    def float(self):
        """Read as float val"""
        return types.Float(self._process.run_script_generic("""send(ptr("{}").readFloat())""".format(hex(self.address)), raw=True, unload=True)[0][0])

    @float.setter
    def float(self, val):
        self._process.run_script_generic("""send(ptr("{}").writeFloat({}))""".format(hex(self.address), val), raw=True, unload=True)
    
    @property
    def pointer(self):
        """Read as pointer val"""
        return types.Pointer(common.auto_int(self._process.run_script_generic("""send(ptr("{}").readPointer())""".format(hex(self.address)), raw=True, unload=True)[0][0]))

    @pointer.setter
    def pointer(self, val):
        common.auto_int(self._process.run_script_generic("""send(ptr("{}").writePointer(ptr("{}")))""".format(hex(self.address), hex(val)), raw=True, unload=True)[0][0])

    @property
    def breakpoint(self):
        """bool: Does this address have an active breakpoint?"""
        return self.address in self._process.memory._active_breakpoints

    @breakpoint.setter
    def breakpoint(self, val):
        """bool: Set this as a breakpoint or remove the breakpoint."""
        
        assert type(val) is bool, "breakpoint set must be boolean."

        # Remove breakpoint
        if val is False:
            # We're already not a breakpoint
            if not self.breakpoint:
                return

            # Remove breakpoint
            self._process.run_script_generic("""ptr("{}").writeS8(1);""".format(hex(self._process.memory._active_breakpoints[self.address])), raw=True, unload=True)
            self._process.memory._active_breakpoints.pop(self.address)

        # Add breakpoint
        else:
            # Breakpoint already exists
            if self.breakpoint:
                return

            unbreak = int(self._process.run_script_generic('generic_suspend_until_true.js', replace={"FUNCTION_HERE": hex(self.address)})[0][0],16)
            #print('Unsuspend pointer: ' + hex(unbreak))
            self._process.memory._active_breakpoints[self.address] = unbreak


    @property
    def bytes(self):
        """bytes: Return this as raw bytes."""
        if self.address_stop is None:
            length = 1 # Default to 1 byte
        else:
            length = self.address_stop - self.address

        return self._process.run_script_generic("""send('array', ptr("{}").readByteArray({}))""".format(hex(self.address), hex(length)), raw=True, unload=True)[1][0]

    @bytes.setter
    def bytes(self, b):
        if type(b) is str:
            logger.warning("Implicitly converting str to bytes.")
            b = b.encode('latin-1')

        if type(b) is not bytes:
            logger.error("Must use type 'bytes' when writing as bytes.")
            return

        # If we know our size, check that we're not overwriting
        if self.size is not None and len(b) > self.size:
            logger.warning("Writing more bytes than it appears is allocated.")

        self._process.run_script_generic("""ptr("{}").writeByteArray({});""".format(
            hex(self.address),
            json.dumps(list(b)),
            ), raw=True, unload=True)

    @property
    def size(self):
        """int: Size of this MemoryBytes. Only valid if it was generated as a slice, alloc or something else that has known size."""
        if self.address_stop is None:
            return None

        return self.address_stop - self.address

    @property
    def address(self):
        """Pointer: Address of this MemoryBytes."""
        return self.__address

    @address.setter
    def address(self, address):
        # Standardize to Pointer
        if type(address) is int:
            address = types.Pointer(address)
        self.__address = address

    @property
    def address_stop(self):
        """Pointer: Stop address of this MemoryBytes."""
        return self.__address_stop

    @address_stop.setter
    def address_stop(self, address):
        # Standardize to Pointer
        if type(address) is int:
            address = types.Pointer(address)
        self.__address_stop = address
    
    @property
    def instruction(self):
        """AssemblyInstruction: Returns an assembly instruction parsed from what is in memory at this location."""
        return AssemblyInstruction(self._process, self.address)

    @property
    def instruction_block(self):
        """AssemblyBlock: Returns an AssemblyBlock starting at this instruction."""
        return AssemblyBlock(self._process, self.address)

#
# Doc Updates
#
MemoryBytes.implementation.__doc__ = MemoryBytes.replace.__doc__


from ..tracer.assembly_instruction import AssemblyInstruction, AssemblyBlock
from ..native_exception import NativeException




import logging
logger = logging.getLogger(__name__)

import collections
from prettytable import PrettyTable
import datetime
from fnmatch import fnmatch

from .. import common, types

class Modules(object):

    def __init__(self, process):
        self._process = process

        # key == module name, value == dict of symbol->address
        self._symbol_to_address = {}

        # key == address, value == symbol
        self._address_to_symbol = {}

        self.__last_update = datetime.datetime(1970,1,1)

    def lookup_symbol(self, symbol):
        """Generically resolve a symbol.
        
        Examples:
            resolve_symbol(":strlen") -> returns address of strlen resolved globally.
            resolve_symbol("a.out:main") -> returns address of main resolved to a.out.
            resolve_symbol(0x12345) -> returns symbol at that address.
        
        """

        # Resolve address to symbol
        if isinstance(symbol, int):
            try:
                return self._address_to_symbol[symbol]
            except KeyError:
                return None

        module, offset, symbol = common.parse_location_string(symbol)

        # First try to resolve with local symbol table
        try:
            return self._symbol_to_address[module][symbol]
        except KeyError:
            pass
        
        # Fall back to asking Frida to resolve it

        replace_vars = {
                "FUNCTION_SYMBOL_HERE": symbol,
                "FUNCTION_MODULE_HERE": module,
                "FUNCTION_OFFSET_HERE": offset,
                }

        return types.Pointer(common.auto_int(self._process.run_script_generic("resolve_location_address.js", replace=replace_vars, unload=True)[0][0]))

    def __iter__(self):
        return self.modules.__iter__()

    def __len__(self):
        return len(self.modules)

    def __repr__(self):
        attrs = ['Modules', str(len(self))]
        return "<{}>".format(' '.join(attrs))

    def __str__(self):
        table = PrettyTable(['name', 'base', 'size', 'path'])

        for module in self:
            table.add_row([module.name, hex(module.base), hex(module.size), module.path])

        table.align['path'] = 'l'
        
        return str(table)

    def __getitem__(self, item):

        # Resolve module by it's name
        if isinstance(item, str):
            return next(mod for mod in self if fnmatch(mod.name, item))

        # Resolve module by the address
        if isinstance(item, int):
            for mod in self:
                if item >= mod.base and item <= mod.base + mod.size:
                    return mod
            return None

        raise NotImplementedError

    @property
    def modules(self):
        """list: Return list of modules."""

        # Time to update the cache
        if datetime.datetime.now() - self.__last_update > datetime.timedelta(seconds=0.5):
            mods = self._process.run_script_generic("""send(Process.enumerateModulesSync());""", raw=True, unload=True)[0][0]
            self.__modules = [Module(self._process, name=mod['name'], base=mod['base'], size=mod['size'], path=mod['path']) for mod in mods]
            self.__last_update = datetime.datetime.now()

        return self.__modules

from .module import Module




import logging
logger = logging.getLogger(__name__)

from termcolor import cprint, colored
import re

from prettytable import PrettyTable

from .. import types, common

class AssemblyBlock(object):
    """Represents an assembly block."""

    def __init__(self, process, address):
        self._process = process
        self.instructions = []
        self.address = address
        self._parse_block()

    def _parse_block(self):
        address = self.address
        end_block = ['ret', 'call', 'branch_relative']

        while True:
            self.instructions.append(AssemblyInstruction(self._process, address))

            for ender in end_block:
                if ender in self.instructions[-1].groups:
                    return
            
            address = self.instructions[-1].address_next

    def __getitem__(self, item):
        return self.instructions[item]

    def __str__(self):
        s = []
        for i in self.instructions:
            s.append(str(i))
        return '\n'.join(s)

    def __repr__(self):
        attrs = ['AssemblyBlock', str(len(self.instructions)), 'instructions']
        return '<' + ' '.join(attrs) + '>'


class AssemblyInstruction(object):
    """Represents an assembly instruction."""

    def __init__(self, process, address):
        """

        Args:
            process: Process object
            address (int): Address for this instruction. Will load from this address.
        """

        self._process = process
        self.address = address

    def _load_from_address(self):
        inst = self._process.run_script_generic("""send(Instruction.parse({}));""".format(self.address.js), raw=True, unload=True)[0][0]

        self.__address_next = types.Pointer(common.auto_int(inst['next']))
        self.__size = inst['size']
        self.__mnemonic = inst['mnemonic']
        self.__args_str = inst['opStr']
        self.__operands = inst['operands']
        self.__registers_read = inst['regsRead']
        self.__registers_written = inst['regsWritten']
        self.__groups = inst['groups']

    def __str__(self):
        return "{address} {mnemonic: <20}{args}".format(
                    address=colored(hex(self.address), attrs=['bold']) + ":",
                    mnemonic=colored(self.mnemonic, "cyan"),
                    args=colored(self.args_str_resolved, "cyan", attrs=['bold']),
                )

    def __repr__(self):
        attrs = ['AssemblyInstruction', hex(self.address), self.mnemonic, self.args_str]
        return '<' + ' '.join(attrs) + '>'

    @property
    def args_str_resolved(self):
        """str: Attempt to resolve addresses in the args str into symbols."""
        s = self.args_str
        
        things = re.findall('0x[0-9a-f]+', s)
        
        for thing in things:
            sym = self._process.modules.lookup_symbol(int(thing,16))
            if sym is not None:
                s = s.replace(thing, str(sym))

        return s

    @property
    def groups(self):
        """list: List of descriptive groups that this instruction belongs to."""
        return self.__groups

    @property
    def registers_written(self):
        """list: List of registers written by this instruction."""
        return self.__registers_written

    @property
    def registers_read(self):
        """list: List of registers that are read by this instruction."""
        return self.__registers_read

    @property
    def operands(self):
        """list: List of operands."""
        return self.__operands

    @property
    def args_str(self):
        """str: Operation arguments as a string."""
        return self.__args_str

    @property
    def mnemonic(self):
        """str: Operation mnemonic."""
        return self.__mnemonic

    @property
    def size(self):
        """int: Size of this instruction in bytes."""
        return self.__size

    @property
    def address_next(self):
        """Pointer: Address of instruction following this one."""
        return self.__address_next

    @property
    def address(self):
        """Pointer: Address where this instruction is located."""
        return self.__address

    @address.setter
    def address(self, address):
        self.__address = types.Pointer(common.auto_int(address))
        self._load_from_address()




import logging
logger = logging.getLogger(__name__)

import colorama
from copy import copy
from termcolor import cprint, colored
from .. import common
from . import ActionFind

class ActionDiffFind:
    """Handle finding things in memory differentially."""


    def __init__(self, process, *args, **kwargs):
        """
        Args:
            process: Parent process instantiation
        """
        self._process = process
        self._args = args
        self._kwargs = kwargs
        self.memory_locations = {}
        self.memory_locations_new = None

    def run(self):
        self.run_menu()

    def run_menu(self):

        menu_options = {
            'n': {'description': 'Search for number', 'action': self.search_number},
            'p': {'description': 'Print current found locations', 'action': self.print_locations},
            'q': {'description': 'Quit', 'action': exit},
        }

        while True:
            
            print("Total memory locations = " + str(len(self.memory_locations)))
            print("")

            for key, info in menu_options.items():
                print("[{key}] {description}".format(key=key, description=info['description']))

            i = input("> ").lower()

            if i not in menu_options:
                logger.error("Invalid choice '{}'.".format(i))
                continue

            menu_options[i]['action']()
            self.merge_results()

    
    def print_locations(self):
        """Prints out what we've found so far."""
        print({hex(x): y for x,y in self.memory_locations.items()})
    
    def merge_results(self):
        """Merge the new results with the existing results."""

        # We didn't actually search
        if self.memory_locations_new is None:
            return

        # If this is the first query, just copy over
        if self.memory_locations == {}:
            self.memory_locations = self.memory_locations_new

        else:
            orig_set = set(self.memory_locations.keys())
            new_set = set(self.memory_locations_new.keys())

            self.memory_locations = {x:self.memory_locations_new[x] for x in orig_set.intersection(new_set)}


        self.memory_locations_new = None

    
    def search_number(self):

        try:
            n = int(input("Number> "), 0)
        except ValueError:
            logger.error("Invalid number")
            return

        kwargs = copy(self._kwargs)

        kwargs['number'] = n
        find = ActionFind(self._process, **kwargs)
        find.run()

        self.memory_locations_new = copy(find.discovered_locations)


    def action_diff(self):

        def find_cb(message, data):
            #print(message)
            found = message['payload']
            
            if found is not None:
                assert type(found) is list, 'Unexpected found type of {}'.format(type(found))
                for f in found:
                    self.discovered_locations[int(f['address'],16)] = pattern_type

        return

        #
        # Actually do search
        #

        for find_pattern, pattern_type in find_patterns.items():

            find_js = self._process.load_js('find_in_memory.js')
            find_js = find_js.replace("SCAN_PATTERN_HERE", find_pattern)

            script = self._process.session.create_script(find_js)
            script.on('message', find_cb)

            logger.debug("Starting Memory find ... ")
            script.load()
